/*
Highly Optimized Object-oriented Many-particle Dynamics -- Blue Edition
(HOOMD-blue) Open Source Software License Copyright 2009-2014 The Regents of
the University of Michigan All rights reserved.

HOOMD-blue may contain modifications ("Contributions") provided, and to which
copyright is held, by various Contributors who have granted The Regents of the
University of Michigan the right to modify and/or distribute such Contributions.

You may redistribute, use, and create derivate works of HOOMD-blue, in source
and binary forms, provided you abide by the following conditions:

* Redistributions of source code must retain the above copyright notice, this
list of conditions, and the following disclaimer both in the code and
prominently in any materials provided with the distribution.

* Redistributions in binary form must reproduce the above copyright notice, this
list of conditions, and the following disclaimer in the documentation and/or
other materials provided with the distribution.

* All publications and presentations based on HOOMD-blue, including any reports
or published results obtained, in whole or in part, with HOOMD-blue, will
acknowledge its use according to the terms posted at the time of submission on:
http://codeblue.umich.edu/hoomd-blue/citations.html

* Any electronic documents citing HOOMD-Blue will link to the HOOMD-Blue website:
http://codeblue.umich.edu/hoomd-blue/

* Apart from the above required attributions, neither the name of the copyright
holder nor the names of HOOMD-blue's contributors may be used to endorse or
promote products derived from this software without specific prior written
permission.

Disclaimer

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND/OR ANY
WARRANTIES THAT THIS SOFTWARE IS FREE OF INFRINGEMENT ARE DISCLAIMED.

IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#ifdef WIN32
#pragma warning( push )
#pragma warning( disable : 4103 4244 )
#endif

#include <iostream>

#include <boost/bind.hpp>
#include <boost/function.hpp>

#include "ConstForceCompute.h"
#include "TwoStepNVT.h"
#include "ComputeThermo.h"
#ifdef ENABLE_CUDA
#include "TwoStepNVTGPU.h"
#endif
#include "IntegratorTwoStep.h"

#include "AllPairPotentials.h"
#include "NeighborList.h"
#include "Initializers.h"

#include <math.h>

using namespace std;
using namespace boost;

/*! \file nvt_updater_test.cc
    \brief Implements unit tests for NVTUpdater and descendants
    \ingroup unit_tests
*/

//! name the boost unit test module
#define BOOST_TEST_MODULE NVTUpdaterTests
#include "boost_utf_configure.h"

//! Reference trace from a mathematica solution to a simple nose-hoover system
double p_reference[]  = { 0. , 0.0037406566640887534 , 0.007462808071668591 ,
                          0.011166655846694415 , 0.014852455068522275 , 0.018520428470662827 , 0.02217079658096173 ,
                          0.025803781175335324 , 0.029419603323072033 , 0.03301848079788623 , 0.036600631307961595 ,
                          0.04016626771538785 , 0.0437156015862913 , 0.047248844515139415 , 0.05076620589105645 ,
                          0.0542678901856463 , 0.05775410132343707 , 0.06122504316215906 , 0.06468091624262627 ,
                          0.06812191842665004 , 0.07154824753120685 , 0.07496009977783834 , 0.07835766718889466 ,
                          0.08174114111093467 , 0.08511071273206282 , 0.08846657011579426 , 0.09180889849464546 ,
                          0.09513788300068277 , 0.0984537115881319 , 0.10175656595890113 , 0.10504662107098113 ,
                          0.10832405187844804 , 0.11158903333537797 , 0.11484174039584709 , 0.11808234801393151 ,
                          0.12131103114370739 , 0.12452796472113263 , 0.12773331985632586 , 0.13092725958345608 ,
                          0.1341099459581678 , 0.13728154103610551 , 0.14044220687291373 , 0.14359210552423685 ,
                          0.14673139904571947 , 0.149860249493006 , 0.15297881875375274 , 0.1560872631051862 ,
                          0.15918573272963144 , 0.1622743775917873 , 0.1653533476563527 , 0.16842279288802642 ,
                          0.17148286325150724 , 0.1745337087114941 , 0.17757547923268574 , 0.18060832441178182 ,
                          0.18363238798589365 , 0.18664780919717636 , 0.18965472719293613 , 0.19265328112047922 ,
                          0.1956436101271119 , 0.19862585336014038 , 0.2016001499668709 , 0.20456663909460965 ,
                          0.2075254591590168 , 0.21047674247562995 , 0.21342061844998642 , 0.2163572164783878 ,
                          0.21928666595713575 , 0.2222090962825319 , 0.2251246368508779 , 0.22803341705847538 ,
                          0.2309355663016259 , 0.23383121272334229 , 0.23672047856070985 , 0.23960348440081297 ,
                          0.24248035083366704 , 0.2453511984492875 , 0.24821614783768972 , 0.2510753195888892 ,
                          0.25392883429290125 , 0.25677681254528084 , 0.25961937541990243 , 0.26245663735372005 ,
                          0.2652887095769494 , 0.26811570331980633 , 0.27093772981250674 , 0.27375490028526644 ,
                          0.2765673259683012 , 0.27937511809182686 , 0.28217838788605926 , 0.28497724658121426 ,
                          0.28777180540750763 , 0.2905621755951552 , 0.29334846694311933 , 0.29613078273311993 ,
                          0.29890922454341684 , 0.3016838939575917 , 0.3044548925592264 , 0.30722232193190263 ,
                          0.3099862836592021 , 0.3127468793247066 , 0.31550421051199773 , 0.3182583788046574 ,
                          0.32100948578626726 , 0.32375763304040905 , 0.3265029215780647 , 0.32924544727555083 ,
                          0.3319853033762342 , 0.3347225831075387 , 0.33745737969688827 , 0.3401897863717068 ,
                          0.3429198963594183 , 0.3456478028874465 , 0.3483735991832155 , 0.3510973784741491 ,
                          0.35381923398767134 , 0.356539258951206 , 0.3592575463677527 , 0.3619741853472208 ,
                          0.3646892618330899 , 0.3674028616897767 , 0.3701150707816978 , 0.37282597497326997 ,
                          0.3755356601289098 , 0.37824421211303416 , 0.38095171679005957 , 0.3836582600244028 ,
                          0.3863639276804805 , 0.38906880562270946 , 0.39177297965696917 , 0.3944765329001641 ,
                          0.3971795448915754 , 0.3998820949422996 , 0.40258426236343314 , 0.40528612646607254 ,
                          0.40798776656131436 , 0.41068926196025507 , 0.41339069197399114 , 0.41609213591361915 ,
                          0.4187936730902356 , 0.4214953828149369 , 0.42419734439375933 , 0.4268996354878481 ,
                          0.4296023300003089 , 0.43230550133581885 , 0.4350092228990552 , 0.4377135680946951 ,
                          0.4404186103274155 , 0.4431244230018938 , 0.44583107952280693 , 0.44853865329483217 ,
                          0.45124721772264664 , 0.4539568462109274 , 0.4566676121643517 , 0.4593795881093839 ,
                          0.4620928429410612 , 0.4648074446409567 , 0.467523461191036 , 0.4702409605732647 ,
                          0.4729600107696085 , 0.4756806797620329 , 0.47840303553250363 , 0.4811271460629862 ,
                          0.4838530793354463 , 0.48658090333184956 , 0.4893106860341615 , 0.4920424964439555 ,
                          0.4947764018763795 , 0.49751246431364954 , 0.5002507455613009 , 0.5029913074248692 ,
                          0.5057342117098896 , 0.5084795202218979 , 0.5112272947664293 , 0.5139775971490194 ,
                          0.5167304891752034 , 0.5194860326505171 , 0.5222442893804957 , 0.5250053211706747 ,
                          0.5277691898265896 , 0.5305359571537759 , 0.533305684957769 , 0.5360784350441042 ,
                          0.5388542691956931 , 0.5416332468919319 , 0.5444154233548733 , 0.54720085334635 ,
                          0.5499895916281953 , 0.552781692962242 , 0.5555772121103232 , 0.5583762038342717 ,
                          0.5611787228959206 , 0.5639848240571028 , 0.5667945620796514 , 0.5696079917253993 ,
                          0.5724251677561796 , 0.5752461449338251 , 0.5780709780201687 , 0.5808997217770437 ,
                          0.5837324309662828 , 0.5865691603497192 , 0.589409964261904 , 0.5922548927049477 ,
                          0.5951039932419585 , 0.5979573134120566 , 0.6008149007543627 , 0.6036768028079975 ,
                          0.6065430671120813 , 0.6094137412057348 , 0.6122888726280783 , 0.6151685089182327 ,
                          0.6180526976153183 , 0.6209414862584558 , 0.6238349223867656 , 0.6267330535393684 ,
                          0.6296359272553845 , 0.6325435910739348 , 0.6354560925341394 , 0.6383734791853581 ,
                          0.6412957965195032 , 0.6442230849049252 , 0.6471553840061389 , 0.65009273348766 ,
                          0.653035173014003 , 0.6559827422496837 , 0.6589354808592168 , 0.6618934285071179 ,
                          0.6648566248579016 , 0.6678251095760837 , 0.6707989223261788 , 0.6737781027727024 ,
                          0.6767626905801696 , 0.6797527254130955 , 0.6827482469359956 , 0.6857492948133844 ,
                          0.6887559087097778 , 0.6917681280013359 , 0.6947859868403071 , 0.6978095159090744 ,
                          0.700838745852957 , 0.7038737073172739 , 0.7069144309473443 , 0.7099609473884871 ,
                          0.7130132872860218 , 0.7160714812852669 , 0.719135560031542 , 0.7222055541701659 ,
                          0.7252814943464577 , 0.7283634112057366 , 0.7314513353933216 , 0.7345452975545318 ,
                          0.7376453283346864 , 0.7407514583791043 , 0.7438637183875781 , 0.7469821365850217 ,
                          0.7501067339693585 , 0.7532375305130846 , 0.7563745461886975 , 0.7595178009686936 ,
                          0.7626673148255697 , 0.7658231077318228 , 0.7689851996599494 , 0.7721536105824466 ,
                          0.7753283604718108 , 0.7785094693005391 , 0.7816969570411281 , 0.7848908436660748 ,
                          0.7880911491478756 , 0.7912978934590277 , 0.7945110965720276 , 0.7977307784593722 ,
                          0.8009569588014012 , 0.8041896497683063 , 0.8074288585343207 , 0.8106745922322212 ,
                          0.8139268579947844 , 0.8171856629547875 , 0.8204510142450068 , 0.8237229189982196 ,
                          0.8270013843472023 , 0.8302864174247321 , 0.8335780253635853 , 0.836876215296539 ,
                          0.8401809943563702 , 0.8434923696758552 , 0.8468103483877713 , 0.8501349376248949 ,
                          0.8534661445200031 , 0.856803976306924 , 0.860148436467478 , 0.8634995180100017 ,
                          0.8668572125718698 , 0.8702215117904559 , 0.8735924073031346 , 0.8769698907472798 ,
                          0.8803539537602659 , 0.8837445879794668 , 0.8871417850422569 , 0.8905455365860102 ,
                          0.8939558342481011 , 0.8973726696659035 , 0.9007960344767917 , 0.9042259203181398 ,
                          0.9076623188273221 , 0.9111052216417125 , 0.9145546203986855 , 0.9180105061742552 ,
                          0.9214728588178506 , 0.924941651435882 , 0.9284168570988808 , 0.9318984488773776 ,
                          0.9353863998419031 , 0.9388806830629883 , 0.942381271611164 , 0.9458881385569612 ,
                          0.9494012569709103 , 0.9529205999235426 , 0.9564461404853887 , 0.9599778517269795 ,
                          0.9635157067188457 , 0.9670596785315184 , 0.9706097402355282 , 0.974165864901406 ,
                          0.9777280248534969 , 0.9812961796228195 , 0.9848702819416935 , 0.988450285113091 ,
                          0.992036142439983 , 0.9956278072253415 , 0.9992252327721376 , 1.002828372383343 ,
                          1.0064371793619296 , 1.0100516070108683 , 1.0136716086331308 , 1.0172971375316888 ,
                          1.0209281470095137 , 1.024564590369577 , 1.0282064209155242 , 1.0318535895609287 ,
                          1.0355060335852333 , 1.0391636859949938 , 1.0428264797967675 , 1.0464943479971098 ,
                          1.050167223602578 , 1.053845039619728 , 1.0575277290551164 , 1.0612152249152995 ,
                          1.064907460206834 , 1.068604367936276 , 1.072305881110182 , 1.0760119327351085 ,
                          1.079722455817612 , 1.083437382805783 , 1.0871566345712367 , 1.0908801219878652 ,
                          1.0946077556417917 , 1.098339446119138 , 1.1020751040060268 , 1.1058146398885802 ,
                          1.1095579643529205 , 1.1133049879851706 , 1.1170556213714522 , 1.120809775097888 ,
                          1.1245673597506003 , 1.1283282859157115 , 1.1320924641793437 , 1.1358598051118731 ,
                          1.1396302123313737 , 1.1434035741598885 , 1.147179776973848 , 1.1509587071496838 ,
                          1.154740251063826 , 1.158524295092706 , 1.1623107256127543 , 1.1660994290004023 ,
                          1.1698902916320804 , 1.1736831998842199 , 1.1774780401332512 , 1.181274698755606 ,
                          1.1850730621277143 , 1.1888730166260078 , 1.1926744458972292 , 1.1964772170274234 ,
                          1.2002811910277134 , 1.2040862289028933 , 1.2078921916577587 , 1.2116989402971032 ,
                          1.2155063358257223 , 1.2193142392484098 , 1.2231225115699609 , 1.2269310137951697 ,
                          1.2307396069288312 , 1.2345481519757397 , 1.2383565099406901 , 1.2421645418284768 ,
                          1.245972107985976 , 1.24977905590862 , 1.253585221414004 , 1.2573904398966882 ,
                          1.2611945467512342 , 1.2649973773722019 , 1.2687987671541525 , 1.272598551491647 ,
                          1.2763965657792453 , 1.2801926454115087 , 1.283986625782998 , 1.2877783422882738 ,
                          1.2915676303218964 , 1.2953543252784272 , 1.2991382625115435 , 1.3029192701033678 ,
                          1.306697160778294 , 1.3104717453311627 , 1.3142428345568165 , 1.3180102392500972 ,
                          1.321773770205847 , 1.325533238218907 , 1.3292884540841192 , 1.333039228596326 ,
                          1.3367853725503693 , 1.3405266967410903 , 1.344263011963331 , 1.3479941290119337 ,
                          1.3517198586817403 , 1.3554400089248115 , 1.3591543728824846 , 1.3628627393017674 ,
                          1.3665648969493578 , 1.370260634591953 , 1.3739497409962511 , 1.3776320049289499 ,
                          1.3813072151567476 , 1.3849751604463414 , 1.3886356295644293 , 1.392288411277709 ,
                          1.3959332943528788 , 1.399570067556636 , 1.4031985196556784 , 1.4068184381246842 ,
                          1.410429596439828 , 1.4140317659261175 , 1.4176247190279683 , 1.4212082281897978 ,
                          1.4247820658560224 , 1.4283460044710592 , 1.4318998164793242 , 1.435443274325234 ,
                          1.4389761504532057 , 1.4424982173076561 , 1.446009247333001 , 1.4495090120024272 ,
                          1.4529972751244484 , 1.45647379848317 , 1.4599383440616454 , 1.4633906738429285 ,
                          1.4668305498100738 , 1.4702577339461351 , 1.4736719882341658 , 1.4770730746572205 ,
                          1.4804607551983529 , 1.4838347918406172 , 1.4871949465670669 , 1.490540979902473 ,
                          1.4938726472971189 , 1.4971897054753525 , 1.50049191132785 , 1.5037790217452869 ,
                          1.5070507936183386 , 1.5103069838376808 , 1.51354734929399 , 1.516771646877941 ,
                          1.5199796334802098 , 1.5231710659914723 , 1.5263457013024044 , 1.5295032944516653 ,
                          1.532643599720209 , 1.5357663751007835 , 1.5388713786531127 , 1.5419583684369198 ,
                          1.5450271025119289 , 1.5480773389378635 , 1.5511088357744482 , 1.554121351081406 ,
                          1.557114642918461 , 1.560088469345337 , 1.563042588421758 , 1.5659767565299487 ,
                          1.5688907352874977 , 1.5717842910329047 , 1.5746571901063273 , 1.577509198847923 ,
                          1.5803400835978485 , 1.5831496106962613 , 1.5859375464833187 , 1.5887036572991782 ,
                          1.5914477094839967 , 1.5941694693779316 , 1.5968687032848028 , 1.5995451771502551 ,
                          1.602198668674698 , 1.6048289600883774 , 1.6074358336215382 , 1.6100190715044262 ,
                          1.6125784559672864 , 1.6151137692403643 , 1.6176247935539054 , 1.620111311138155 ,
                          1.6225731042233584 , 1.625009955039761 , 1.6274216456749586 , 1.6298079610736789 ,
                          1.6321687034245878 , 1.6345036786011116 , 1.6368126924766757 , 1.6390955509247063 ,
                          1.6413520598186286 , 1.6435820250318685 , 1.6457852524378516 , 1.6479615479100036 ,
                          1.6501107173217504 , 1.6522325665465174 , 1.6543269012725377 , 1.6563935351848516 ,
                          1.6584323026778973 , 1.6604430407058024 , 1.6624255862226942 , 1.6643797761827006 ,
                          1.6663054475399488 , 1.6682024372485662 , 1.6700705822626807 , 1.6719097195364196 ,
                          1.6737196860239107 , 1.675500318679281 , 1.6772514545281871 , 1.67897294518941 ,
                          1.6806646638767864 , 1.6823264852762971 , 1.6839582840739231 , 1.6855599349556454 ,
                          1.687131312607445 , 1.6886722917153028 , 1.6901827469651995 , 1.6916625530431166 ,
                          1.6931115846350344 , 1.6945297164269344 , 1.6959168240253002 , 1.6972728046974213 ,
                          1.6985975756123426 , 1.6998910545874233 , 1.7011531594400215 , 1.702383807987496 ,
                          1.7035829180472053 , 1.7047504074365083 , 1.7058861939727634 , 1.706990195473329 ,
                          1.7080623297555644 , 1.7091025146368275 , 1.710110670578832 , 1.7110867459110493 ,
                          1.712030705154604 , 1.7129425130089628 , 1.7138221341735915 , 1.714669533347956 ,
                          1.715484675231522 , 1.7162675245237562 , 1.717018045924124 , 1.7177362041320914 ,
                          1.7184219638471248 , 1.7190752897686896 , 1.7196961520220528 , 1.7202845525112838 ,
                          1.720840504616172 , 1.7213640217305444 , 1.7218551172482273 , 1.7223138045630473 ,
                          1.7227400970688311 , 1.7231340081594053 , 1.7234955512285965 , 1.7238247396702313 ,
                          1.724121586878136 , 1.7243861062468018 , 1.7246183204311771 , 1.7248182842725521 ,
                          1.7249860595884199 , 1.7251217081962735 , 1.7252252919136064 , 1.7252968725579114 ,
                          1.7253365119466821 , 1.7253442718974117 , 1.7253202142275932 , 1.7252644007547198 ,
                          1.7251768932962848 , 1.7250577537396394 , 1.72490705760077 , 1.724724909218315 ,
                          1.7245114164971838 , 1.7242666873422845 , 1.723990829658526 , 1.723683951350817 ,
                          1.723346160324066 , 1.722977564483182 , 1.7225782717330733 , 1.722148389978649 ,
                          1.7216880271248172 , 1.7211972914494518 , 1.7206763085888923 , 1.7201252269532132 ,
                          1.7195441963108813 , 1.7189333664303625 , 1.718292887080124 , 1.7176229080286316 ,
                          1.7169235790443524 , 1.716195049895753 , 1.7154374703512993 , 1.7146509901794582 ,
                          1.7138357591486961 , 1.712991928026082 , 1.71211966694044 , 1.711219161538218 ,
                          1.710290597656559 , 1.7093341611326067 , 1.7083500378035044 , 1.7073384135063951 ,
                          1.7062994740784225 , 1.70523340535673 , 1.7041403931784609 , 1.7030206233807583 ,
                          1.7018742818007657 , 1.7007015561785508 , 1.6995026532028763 , 1.6982777880616406 ,
                          1.6970271757126885 , 1.695751031113864 , 1.6944495692230115 , 1.6931230049979755 ,
                          1.6917715533966 , 1.6903954293767298 , 1.688994847896209 , 1.6875700239128815 ,
                          1.6861211723845921 , 1.6846485111723002 , 1.6831522741184786 , 1.6816326979051435 ,
                          1.6800900189820691 , 1.6785244737990301 , 1.6769362988058008 , 1.6753257304521554 ,
                          1.6736930051878687 , 1.6720383594627144 , 1.6703620297264674 , 1.668664252428902 ,
                          1.6669452640197926 , 1.6652053046556499 , 1.6634446253882678 , 1.6616634764246947 ,
                          1.6598621078748614 , 1.658040769848698 , 1.6561997124561358 , 1.6543391858071055 ,
                          1.6524594400115376 , 1.6505607251793624 , 1.6486432914205114 , 1.6467073888449146 ,
                          1.6447532675625032 , 1.6427811816512323 , 1.6407913897819377 , 1.638784148135406 ,
                          1.6367597128862688 , 1.6347183402091583 , 1.6326602862787063 , 1.6305858072695452 ,
                          1.6284951593563068 , 1.626388598713623 , 1.624266381516126 , 1.6221287639384478 ,
                          1.6199760021788387 , 1.6178083556971423 , 1.615626082346258 , 1.613429437373496 ,
                          1.6112186760261655 , 1.6089940535515759 , 1.6067558251970364 , 1.6045042462098569 ,
                          1.602239571837347 , 1.5999620573268156 , 1.5976719579255725 , 1.5953695288809273 ,
                          1.5930550255504876 , 1.590728704469092 , 1.5883908159257782 , 1.5860416081260946 ,
                          1.5836813292755885 , 1.5813102275798079 , 1.5789285512443003 , 1.5765365484746137 ,
                          1.5741344674762958 , 1.5717225564548944 , 1.5693010636159572 , 1.5668702371650318 ,
                          1.5644303254655456 , 1.5619815748763552 , 1.5595242228734338 , 1.5570585055574702 ,
                          1.5545846590291532 , 1.5521029193891718 , 1.5496135227382146 , 1.5471167051769705 ,
                          1.5446127028061285 , 1.5421017517263775 , 1.5395840880384064 , 1.5370599478429037 ,
                          1.534529567271815 , 1.5319931767306003 , 1.529450997025246 , 1.5269032482173317 ,
                          1.5243501503684385 , 1.5217919235401458 , 1.519228787794034 , 1.5166609631916832 ,
                          1.5140886697946732 , 1.5115121276645846 , 1.5089315568629973 , 1.5063471774514916 ,
                          1.5037592090992236 , 1.5011678621353293 , 1.4985733380805788 , 1.4959758381460255 ,
                          1.4933755635427222 , 1.4907727154817223 , 1.4881674951740793 , 1.4855601038308464 ,
                          1.4829507426630764 , 1.4803396128818231 , 1.4777269156981394 , 1.4751128523230788 ,
                          1.4724976227722693 , 1.469881414816732 , 1.4672644091389062 , 1.464646786339043 ,
                          1.462028727017393 , 1.4594104117742082 , 1.456792021209739 , 1.4541737359242366 ,
                          1.4515557365179523 , 1.448938203591137 , 1.446321317744042 , 1.4437052595769182 ,
                          1.4410902072852079 , 1.4384763250795751 , 1.4358637721294 , 1.4332527075973398 ,
                          1.4306432906460522 , 1.4280356804381948 , 1.4254300361364256 , 1.4228265169034022 ,
                          1.4202252819017822 , 1.4176264902942233 , 1.4150303012433834 , 1.412436873911868 ,
                          1.4098463634716278 , 1.4072589107946472 , 1.4046746535679957 , 1.4020937294787432 ,
                          1.3995162762139595 , 1.3969424314607135 , 1.3943723329060755 , 1.391806118237115 ,
                          1.389243925140902 , 1.3866858913045055 , 1.3841321544149952 , 1.3815828521383922 ,
                          1.3790381163494154 , 1.3764980656667187 , 1.3739628168972744 , 1.3714324868480556 ,
                          1.3689071923260345 , 1.3663870501381838 , 1.363872177091476 , 1.361362689992884 ,
                          1.3588587056493797 , 1.356360340867936 , 1.3538677124555256 , 1.351380937098499 ,
                          1.3489001240191008 , 1.3464253711235155 , 1.343956775414588 , 1.341494433895164 ,
                          1.3390384435680887 , 1.3365889014362071 , 1.3341459045023645 , 1.3317095497694065 ,
                          1.329279934240178 , 1.326857154917524 , 1.3244413088042903 , 1.322032492567172 ,
                          1.3196307941430157 , 1.3172362925139318 , 1.3148490662935894 , 1.3124691940956568 ,
                          1.3100967545338034 , 1.3077318262216975 , 1.305374487773008 , 1.3030248178014037 ,
                          1.3006828949205533 , 1.298348797744125 , 1.2960226048857888 , 1.2937043948367168 ,
                          1.2913942385779376 , 1.2890921975529248 , 1.286798332671515 , 1.2845127048435443 ,
                          1.2822353749788487 , 1.2799664039872638 , 1.277705852778627 , 1.2754537822627732 ,
                          1.2732102533495393 , 1.270975326948761 , 1.2687490639702745 , 1.2665315253239162 ,
                          1.2643227718147596 , 1.2621228577410748 , 1.2599318280264555 , 1.2577497269683682 ,
                          1.255576598864277 , 1.2534124880116473 , 1.2512574387079438 , 1.2491114952506317 ,
                          1.246974701937176 , 1.2448471030650416 , 1.2427287429316933 , 1.2406196658345972 ,
                          1.2385199160712173 , 1.2364295379390187 , 1.2343485728235346 , 1.232277052435432 ,
                          1.230215006709653 , 1.2281624655821552 , 1.2261194589888964 , 1.224086016865835 ,
                          1.2220621691489288 , 1.2200479457741353 , 1.2180433766774126 , 1.2160484917947185 ,
                          1.214063321062011 , 1.2120878944152478 , 1.2101222417903867 , 1.2081663921202508 ,
                          1.2062203664579438 , 1.2042841825061323 , 1.2023578579856504 , 1.2004414106173327 ,
                          1.1985348581220134 , 1.1966382182205273 , 1.1947515086337086 , 1.1928747470823915 ,
                          1.191007951287411 , 1.1891511389696012 , 1.1873043278497966 , 1.1854675356488318 ,
                          1.1836407798696262 , 1.1818240728029692 , 1.1800174220446913 , 1.1782208351021282 ,
                          1.1764343194826155 , 1.1746578826934888 , 1.1728915322420836 , 1.1711352756357354 ,
                          1.1693891203817797 , 1.167653073987552 , 1.1659271439603878 , 1.1642113378076224 ,
                          1.1625056630365918 , 1.1608101271464568 , 1.1591247349306126 , 1.1574494859315245 ,
                          1.155784379268994 , 1.1541294140628215 , 1.1524845894328086 , 1.1508499044987561 ,
                          1.149225358380465 , 1.147610950197737 , 1.1460066790703725 , 1.144412544118173 ,
                          1.1428285444609392 , 1.1412546792184723 , 1.1396909475105734 , 1.1381373474343746 ,
                          1.1365938723765343 , 1.1350605146874788 , 1.1335372667340593 , 1.1320241208831276 ,
                          1.1305210695015353 , 1.1290281049561335 , 1.1275452196137743 , 1.1260724058413087 ,
                          1.1246096560055885 , 1.1231569624734654 , 1.121714317611791 , 1.1202817137874166 ,
                          1.1188591430889514 , 1.1174465942859235 , 1.116044054387794 , 1.1146515104382366 ,
                          1.113268949480924 , 1.1118963585595287 , 1.1105337247177247 , 1.1091810349991844 ,
                          1.1078382764475811 , 1.1065054361065876 , 1.1051825010198768 , 1.1038694582311221 ,
                          1.1025662947839963 , 1.1012729977707192 , 1.0999895542973512 , 1.098715947720335 ,
                          1.0974521603850396 , 1.096198174636834 , 1.0949539728210873 , 1.0937195372831685 ,
                          1.0924948503684468 , 1.0912798944222908 , 1.09007465179007 , 1.0888791048171533 ,
                          1.0876932358489098 , 1.0865170272307083 , 1.0853504613079181 , 1.084193520425908 ,
                          1.0830461869300474 , 1.081908443165705 , 1.0807802714782497 , 1.0796616541887447 ,
                          1.0785525715131548 , 1.0774530017857054 , 1.0763629236419374 , 1.0752823157173907 ,
                          1.074211156647606 , 1.0731494250681237 , 1.0720970996144836 , 1.0710541589222267 ,
                          1.070020581626893 , 1.0689963463640233 , 1.0679814317691574 , 1.066975816477836 ,
                          1.0659794791255992 , 1.0649923983479874 , 1.0640145527805411 , 1.0630459210588008 ,
                          1.0620864818183064 , 1.061136213690187 , 1.0601950940097127 , 1.0592630988253422 ,
                          1.0583402045090322 , 1.057426387432739 , 1.0565216239684192 , 1.055625890488029 ,
                          1.0547391633635252 , 1.053861418966864 , 1.052992633670002 , 1.0521327838448957 ,
                          1.0512818458635014 , 1.0504397960977756 , 1.0496066109196749 , 1.0487822667011553 ,
                          1.0479667398141739 , 1.0471600066306868 , 1.0463620435226506 , 1.0455728268620217 ,
                          1.0447923322814665 , 1.0440205346286513 , 1.0432574090988973 , 1.042502930887526 ,
                          1.041757075189859 , 1.0410198172012177 , 1.0402911321169235 , 1.039570995132298 ,
                          1.0388593814426628 , 1.0381562662433392 , 1.0374616247296489 , 1.0367754320969131 ,
                          1.0360976635404533 , 1.0354282942555915 , 1.0347672994376484 , 1.0341146542819462 ,
                          1.033470333983806 , 1.0328343137385494 , 1.0322065683560209 , 1.0315870722731366 ,
                          1.030975800290636 , 1.0303727272115473 , 1.0297778278388978 , 1.0291910769757162 ,
                          1.02861244942503 , 1.0280419199898674 , 1.027479463473256 , 1.0269250546782238 ,
                          1.0263786684077991 , 1.0258402794650094 , 1.0253098626528832 , 1.024787392774448 ,
                          1.0242728446327318 , 1.0237661930307627 , 1.0232674127715686 , 1.0227764786581772 ,
                          1.0222933653179445 , 1.0218180473293559 , 1.021350499637541 , 1.0208906971939924 ,
                          1.0204386149502023 , 1.0199942278576628 , 1.0195575108678665 , 1.0191284389323056 ,
                          1.0187069870024725 , 1.0182931300298594 , 1.017886842965959 , 1.0174881007622634 ,
                          1.0170968783702647 , 1.0167131507414555 , 1.0163368928273282 , 1.015968079579375 ,
                          1.0156066859490882 , 1.01525268688796 , 1.014906057286953 , 1.014566772216904 ,
                          1.0142348071131189 , 1.0139101374206723 , 1.0135927385846386 , 1.0132825860500927 ,
                          1.0129796552621089 , 1.012683921665762 , 1.0123953607061262 , 1.0121139478282764 ,
                          1.011839658477287 , 1.011572468098233 , 1.0113123521361884 , 1.011059286036228 ,
                          1.0108132452434264 , 1.0105742052028583 , 1.010342141359598 , 1.0101170291587205 ,
                          1.0098988440422532 , 1.0096875617739949 , 1.0094831584736141 , 1.0092856102721366 ,
                          1.0090948933005885 , 1.0089109836899959 , 1.0087338575713842 , 1.00856349107578 ,
                          1.0083998603342088 , 1.008242941477697 , 1.0080927106372701 , 1.0079491439439545 ,
                          1.0078122175287758 , 1.0076819075227603 , 1.0075581900569337 , 1.007441041262322 ,
                          1.0073304372699514 , 1.0072263542108475 , 1.0071287681288434 , 1.0070376551856535 ,
                          1.0069529923039076 , 1.0068747564606888 , 1.0068029246330799 , 1.0067374737981636 ,
                          1.0066783809330226 , 1.0066256230147401 , 1.0065791770203985 , 1.0065390199270807 ,
                          1.0065051287118696 , 1.0064774803518477 , 1.006456051824098 , 1.0064408201057031 ,
                          1.006431762173746 , 1.0064288550053093 , 1.006432075577476 , 1.0064414008673288 ,
                          1.0064568078519502 , 1.0064782735084234 , 1.006505774813831 , 1.006539288745256 ,
                          1.006578792318982 , 1.0066242630607594 , 1.0066756788338675 , 1.006733017505867 ,
                          1.0067962569443183 , 1.0068653750167813 , 1.0069403495908165 , 1.0070211585339843 ,
                          1.007107779713845 , 1.007200190997959 , 1.0072983702538867 , 1.0074022953491883 ,
                          1.0075119441514242 , 1.0076272945281546 , 1.00774832434694 , 1.0078750114753408 ,
                          1.0080073337809172 , 1.0081452691312296 , 1.0082887953938384 , 1.0084378904363038 ,
                          1.0085925321261864 , 1.0087526983310462 , 1.0089183669622537 , 1.0090895164325713 ,
                          1.0092661253688693 , 1.009448172384365 , 1.009635636092276 , 1.0098284951058198 ,
                          1.010026728038214 , 1.0102303135026762 , 1.0104392301124236 , 1.0106534564806742 ,
                          1.010872971220645 , 1.0110977529455538 , 1.0113277802686182 , 1.0115630318030555 ,
                          1.0118034861620833 , 1.0120491219589192 , 1.0122999178067806 , 1.0125558523188851 ,
                          1.01281690410845 , 1.0130830517886933 , 1.013354273972832 , 1.013630549274084 ,
                          1.013911856343683 , 1.0141981742742783 , 1.0144894822411816 , 1.0147857593839105 ,
                          1.0150869848419821 , 1.0153931377549141 , 1.0157041972622238 , 1.0160201425034285 ,
                          1.0163409526180458 , 1.0166666067455927 , 1.0169970840255869 , 1.0173323635975458 ,
                          1.0176724246009865 , 1.0180172461754267 , 1.0183668074603838 , 1.0187210875953747 ,
                          1.0190800657199173 , 1.019443720973529 , 1.0198120324957267 , 1.0201849794260283 ,
                          1.020562540903951 , 1.0209446960690118 , 1.0213314240896292 , 1.0217227044900765 ,
                          1.0221185167380504 , 1.0225188402386014 , 1.0229236543967792 , 1.0233329386176346 ,
                          1.023746672306218 , 1.0241648348675794 , 1.0245874057067692 , 1.025014364228838 ,
                          1.0254456898388358 , 1.025881361941813 , 1.0263213599428198 , 1.0267656632469069 ,
                          1.0272142512591245 , 1.0276671033845226 , 1.0281241990281518 , 1.0285855175950624 ,
                          1.0290510384903049 , 1.0295207411189293 , 1.029994604885986 , 1.0304726091965253 ,
                          1.0309547334750502 , 1.0314409574030077 , 1.03193126046514 , 1.0324256220534311 ,
                          1.0329240215598645 , 1.033426438376424 , 1.0339328518950939 , 1.0344432415078573 ,
                          1.0349575866066985 , 1.035475866583601 , 1.0359980608305488 , 1.0365241487395258 ,
                          1.0370541097025157 , 1.037587923111502 , 1.0381255683584687 , 1.0386670248354 ,
                          1.039212271934279 , 1.03976128904709 , 1.0403140555658166 , 1.0408705508824425 ,
                          1.041430754388952 , 1.0419946454773283 , 1.0425622035510755 , 1.0431334081713202 ,
                          1.0437082385681613 , 1.0442866738468946 , 1.0448686931128146 , 1.0454542754712168 ,
                          1.046043400027396 , 1.0466360458866473 , 1.0472321921542656 , 1.0478318179355464 ,
                          1.0484349023357842 , 1.0490414244602746 , 1.0496513634143123 , 1.0502646983031927 ,
                          1.0508814082322107 , 1.0515014723066611 , 1.0521248696318393 , 1.0527515793130402 ,
                          1.0533815804555589 , 1.0540148521646904 , 1.05465137354573 , 1.0552911237039726 ,
                          1.055934081750384 , 1.056580226858965 , 1.0572295377534242 , 1.0578819930025967 ,
                          1.0585375711753169 , 1.0591962508404196 , 1.0598580105667397 , 1.0605228289231121 ,
                          1.0611906844783712 , 1.061861555801352 , 1.062535421460889 , 1.0632122600258174 ,
                          1.0638920500649713 , 1.0645747701471862 , 1.0652603988412963 , 1.0659489147161367 ,
                          1.066640296340542 , 1.0673345222833468 , 1.0680315711133863 , 1.0687314213994947 ,
                          1.0694340517105072 , 1.0701394406152582 , 1.070847566694644 , 1.0715584091514965 ,
                          1.0722719462885812 , 1.072988155800355 , 1.0737070153812738 , 1.0744285027257945 ,
                          1.0751525955283732 , 1.0758792714834666 , 1.0766085082855312 , 1.0773402836290231 ,
                          1.078074575208399 , 1.0788113607181158 , 1.0795506178526293 , 1.0802923243063962 ,
                          1.081036457773873 , 1.0817829959495158 , 1.0825319165277816 , 1.0832831972031265 ,
                          1.084036815670007 , 1.08479274962288 , 1.0855509767562015 , 1.0863114747644278 ,
                          1.087074221342016 , 1.0878391941834218 , 1.0886063709831022 , 1.0893757294355133 ,
                          1.090147247235112 , 1.0909209020763544 , 1.091696671653697 , 1.092474533667309 ,
                          1.093254465597309 , 1.0940364440883341 , 1.0948204456380357 , 1.0956064467440656 ,
                          1.0963944239040762 , 1.0971843536157189 , 1.097976212376646 , 1.0987699766845096 ,
                          1.099565623036961 , 1.1003631279316526 , 1.1011624678662362 , 1.1019636193383637 ,
                          1.102766558845687 , 1.1035712628858583 , 1.1043777079565291 , 1.105185870555352 ,
                          1.1059957271799783 , 1.10680725432806 , 1.107620428497249 , 1.1084352261851977 ,
                          1.1092516238895578 , 1.110069598107981 , 1.1108891253381192 , 1.1117101820776247 ,
                          1.1125327448241495 , 1.113356790075345 , 1.1141822943288635 , 1.1150092340823567 ,
                          1.115837585807489 , 1.1166673253620727 , 1.117498428024903 , 1.1183308690571656 ,
                          1.1191646237200468 , 1.1199996672747319 , 1.1208359749824075 , 1.1216735221042595 ,
                          1.122512283901474 , 1.1233522356352368 , 1.124193352566734 , 1.1250356099571517 ,
                          1.1258789830676759 , 1.1267234471594925 , 1.1275689774937876 , 1.1284155493317471 ,
                          1.129263137934557 , 1.1301117185634035 , 1.1309612664794726 , 1.1318117569439503 ,
                          1.1326631652180223 , 1.1335154665628753 , 1.1343686362396945 , 1.1352226495096664 ,
                          1.1360774816339767 , 1.1369331078738119 , 1.1377895034903576 , 1.1386466437448 ,
                          1.139504503898325 , 1.1403630590888807 , 1.1412222835422932 , 1.1420821514257473 ,
                          1.14294263695605 , 1.1438037143500073 , 1.144665357824426 , 1.1455275415961124 ,
                          1.1463902398818728 , 1.1472534268985142 , 1.1481170768628426 , 1.1489811639916643 ,
                          1.1498456625017863 , 1.1507105466100145 , 1.1515757905331558 , 1.1524413684880164 ,
                          1.1533072546914027 , 1.1541734233601215 , 1.1550398487109792 , 1.155906504960782 ,
                          1.1567733663263362 , 1.1576404070244486 , 1.1585076012719258 , 1.159374923285574 ,
                          1.1602423472821994 , 1.1611098474786088 , 1.161977398091609 , 1.1628449733380057 ,
                          1.163712547434606 , 1.1645800945982157 , 1.1654475887383764 , 1.1663150028477913 ,
                          1.167182310496664 , 1.1680494853118253 , 1.1689165009201052 , 1.1697833309483345 ,
                          1.1706499490233435 , 1.171516328771963 , 1.1723824438210233 , 1.173248267797355 ,
                          1.1741137743277887 , 1.1749789370391548 , 1.175843729558284 , 1.1767081255120064 ,
                          1.177572098527153 , 1.1784356222305539 , 1.17929867024904 , 1.1801612162094415 ,
                          1.181023233738589 , 1.1818846964633132 , 1.1827455780104443 , 1.1836058520068133 ,
                          1.1844654920792501 , 1.1853244718545857 , 1.1861827649596504 , 1.1870403450212748 ,
                          1.1878971856662892 , 1.1887532605215245 , 1.1896085432138108 , 1.190463006803016 ,
                          1.1913166238944284 , 1.1921693682186327 , 1.1930212135344749 , 1.1938721336008016 ,
                          1.194722102176459 , 1.195571093020293 , 1.1964190798911505 , 1.1972660365478773 ,
                          1.1981119367493203 , 1.1989567542543251 , 1.1998004628217382 , 1.2006430362104061 ,
                          1.2014844481791749 , 1.202324672486891 , 1.2031636828924006 , 1.2040014531545502 ,
                          1.2048379570321859 , 1.205673168284154 , 1.2065070606693005 , 1.2073396079464722 ,
                          1.2081707838745153 , 1.2090005622122757 , 1.2098289167186003 , 1.2106558211523348 ,
                          1.211481249272326 , 1.2123051748374198 , 1.2131275716064627 , 1.2139484133383007 ,
                          1.214767672956596 , 1.2155853239292247 , 1.2164013411077557 , 1.2172156993458438 ,
                          1.2180283734971435 , 1.2188393384153098 , 1.2196485689539966 , 1.220456039966859 ,
                          1.2212617263075516 , 1.2220656028297288 , 1.2228676443870454 , 1.2236678258331561 ,
                          1.2244661220217155 , 1.2252625078063781 , 1.2260569580407985 , 1.2268494475786316 ,
                          1.2276399512735316 , 1.2284284439791535 , 1.2292149005491517 , 1.229999295837181 ,
                          1.230781604696896 , 1.2315618019819512 , 1.2323398625460014 , 1.233115761242701 ,
                          1.2338894729257048 , 1.2346609724486672 , 1.2354302346652433 , 1.2361972344290872 ,
                          1.2369619465834962 , 1.2377243450364541 , 1.2384844055894482 , 1.2392421053268983 ,
                          1.2399974213332254 , 1.2407503306928498 , 1.241500810490192 , 1.2422488378096723 ,
                          1.2429943897357116 , 1.2437374433527304 , 1.244477975745149 , 1.2452159639973877 ,
                          1.245951385193867 , 1.246684216419008 , 1.2474144347572305 , 1.2481420172929554 ,
                          1.248866941110603 , 1.249589183294594 , 1.2503087209293489 , 1.251025531099288 ,
                          1.2517395908888316 , 1.2524508773824006 , 1.2531593676644153 , 1.2538650388192962 ,
                          1.254567867931464 , 1.255267832085339 , 1.2559649083653417 , 1.2566590738558927 ,
                          1.2573503056414121 , 1.2580385808718058 , 1.258723879029065 , 1.259406180520203 ,
                          1.260085465304736 , 1.2607617133421785 , 1.2614349045920463 , 1.2621050190138547 ,
                          1.262772036567119 , 1.263435937211354 , 1.2640967009060753 , 1.2647543076107983 ,
                          1.2654087372850384 , 1.2660599698883106 , 1.2667079853801304 , 1.2673527637200133 ,
                          1.2679942848674741 , 1.2686325287820281 , 1.2692674754231912 , 1.2698991047504782 ,
                          1.2705273967234043 , 1.2711523313014852 , 1.2717738884367444 , 1.2723920483640103 ,
                          1.2730067935673293 , 1.2736181071495856 , 1.274225972213663 , 1.2748303718624454 ,
                          1.2754312891988169 , 1.2760287073256613 , 1.2766226093458624 , 1.2772129783623043 ,
                          1.2777997974778708 , 1.2783830497954458 , 1.2789627184179133 , 1.2795387864481573 ,
                          1.2801112369890615 , 1.2806800531435099 , 1.2812452180143865 , 1.281806714704575 ,
                          1.2823645263169596 , 1.282918635954424 , 1.2834690267198523 , 1.2840156817145139 ,
                          1.284558584366741 , 1.2850977206327943 , 1.2856330772848734 , 1.286164641095178 ,
                          1.2866923988359076 , 1.2872163372792622 , 1.2877364431974412 , 1.2882527033626443 ,
                          1.2887651045470712 , 1.2892736335229213 , 1.2897782770623947 , 1.2902790219376907 ,
                          1.290775854921009 , 1.2912687627845498 , 1.2917577323005118 , 1.2922427502410951 ,
                          1.2927238033784996 , 1.2932008784849247 , 1.29367396233257 , 1.2941430416936353 ,
                          1.29460810334032 , 1.295069134464931 , 1.295526124998733 , 1.295979065838668 ,
                          1.2964279478817056 , 1.2968727620248153 , 1.2973134991649673 , 1.2977501501991313 ,
                          1.298182706024277 , 1.2986111575373742 , 1.2990354956353927 , 1.299455711215302 ,
                          1.2998717951740721 , 1.300283738408673 , 1.3006915318160743 , 1.3010951662932455 ,
                          1.3014946327371568 , 1.3018899220447777 , 1.3022810251130783 , 1.3026679328390278 ,
                          1.3030506361195966 , 1.303429125851754 , 1.3038033933632778 , 1.304173432839059 ,
                          1.3045392395868878 , 1.3049008089163185 , 1.305258136136906 , 1.3056112165582046 ,
                          1.3059600454897686 , 1.306304618241153 , 1.306644930121912 , 1.3069809764416005 ,
                          1.3073127525097725 , 1.3076402536359832 , 1.3079634751297866 , 1.3082824123007375 ,
                          1.3085970604583903 , 1.3089074149123 , 1.3092134709720202 , 1.3095152239471062 ,
                          1.3098126691471124 , 1.3101058018815932 , 1.3103946174601033 , 1.310679111616549 ,
                          1.3109592829868615 , 1.3112351314211792 , 1.3115066567737335 , 1.3117738588987562 ,
                          1.3120367376504787 , 1.3122952928831322 , 1.3125495244509484 , 1.312799432208159 ,
                          1.3130450160089953 , 1.3132862757076886 , 1.3135232111584705 , 1.3137558222155727 ,
                          1.3139841087332265 , 1.3142080705656636 , 1.3144277075671151 , 1.3146430195918128 ,
                          1.3148540064939882 , 1.3150606681278725 , 1.3152630043476974 , 1.3154610150076944 ,
                          1.3156547003426518 , 1.3158440634266886 , 1.3160291085941818 , 1.3162098401838394 ,
                          1.3163862625343696 , 1.3165583799844807 , 1.31672619687288 , 1.3168897175382765 ,
                          1.3170489463193775 , 1.3172038875548915 , 1.3173545455835263 , 1.31750092474399 ,
                          1.3176430293749908 , 1.3177808638152364 , 1.3179144324034353 , 1.3180437394782951 ,
                          1.318168789378524 , 1.3182895864428306 , 1.318406135009922 , 1.3185184394185068 ,
                          1.318626504007293 , 1.318730333437941 , 1.3188299350645374 , 1.3189253174783924 ,
                          1.319016489269258 , 1.3191034590268866 , 1.3191862353410304 , 1.3192648268014415 ,
                          1.319339241997872 , 1.319409489520074 , 1.3194755779578 , 1.3195375159008018 ,
                          1.319595311938832 , 1.3196489746616424 , 1.3196985126589853 , 1.319743934520613 ,
                          1.3197852488362776 , 1.319822464195731 , 1.3198555891887258 , 1.319884632405014 ,
                          1.3199096024343477 , 1.3199305078664794 , 1.3199473576210234 , 1.3199601632921552 ,
                          1.319968937304734 , 1.3199736920288088 , 1.3199744398344269 , 1.319971193091636 ,
                          1.3199639641704846 , 1.3199527654410204 , 1.319937609273291 , 1.319918508037345 ,
                          1.3198954741032296 , 1.3198685198409932 , 1.3198376576206834 , 1.3198028998123483 ,
                          1.3197642587860356 , 1.3197217469117937 , 1.31967537655967 , 1.3196251600997129 ,
                          1.31957110990197 , 1.3195132383364891 , 1.319451557825134 , 1.319386082434311 ,
                          1.3193168277470253 , 1.3192438093388121 , 1.319167042785207 , 1.3190865436617452 ,
                          1.3190023275439617 , 1.3189144100073926 , 1.3188228066275727 , 1.3187275329800372 ,
                          1.318628604640322 , 1.318526037183962 , 1.3184198461864929 , 1.3183100472234495 ,
                          1.3181966558703675 , 1.3180796877027825 , 1.3179591582962291 , 1.3178350832262435 ,
                          1.3177074780683604 , 1.3175763583981155 , 1.3174417397913378 , 1.317303638748007 ,
                          1.3171620734668157 , 1.3170170621394357 , 1.3168686229575388 , 1.3167167741127965 ,
                          1.3165615337968808 , 1.316402920201463 , 1.3162409515182152 , 1.316075645938809 ,
                          1.3159070216549162 , 1.3157350968582082 , 1.315559889740357 , 1.3153814184930344 ,
                          1.315199701307912 , 1.3150147563766612 , 1.314826601890954 , 1.3146352560424623 ,
                          1.3144407370228575 , 1.3142430630238113 , 1.3140422522369957 , 1.3138383232412407 ,
                          1.313631296192497 , 1.3134211912783185 , 1.3132080286621197 , 1.3129918285073148 ,
                          1.3127726109773183 , 1.3125503962355447 , 1.3123252044454081 , 1.312097055770323 ,
                          1.311865970373704 , 1.311631968418965 , 1.311395070069521 , 1.311155295488786 ,
                          1.3109126648401743 , 1.3106671982871008 , 1.3104189159929793 , 1.3101678381212245 ,
                          1.309913984835251 , 1.3096573762984727 , 1.3093980326743042 , 1.3091359742444655 ,
                          1.3088712224640289 , 1.3086037988991892 , 1.308333725020093 , 1.3080610222968865 ,
                          1.307785712199716 , 1.3075078161987284 , 1.3072273557640695 , 1.3069443523658857 ,
                          1.3066588274743234 , 1.3063708025595293 , 1.3060802990916496 , 1.3057873385408303 ,
                          1.3054919423772182 , 1.3051941320709597 , 1.3048939290922008 , 1.3045913549110881 ,
                          1.3042864309977682 , 1.303979178822387 , 1.303669619855091 , 1.3033577755808166 ,
                          1.303043668201687 , 1.302727320069352 , 1.3024087533100253 , 1.3020879900499216 ,
                          1.3017650524152549 , 1.3014399625322397 , 1.3011127425270903 , 1.3007834145260206 ,
                          1.3004520006552454 , 1.3001185230409786 , 1.2997830038094347 , 1.299445465086828 ,
                          1.2991059289993725 , 1.2987644176732827 , 1.2984209532347728 , 1.2980755578100571 ,
                          1.29772825352535 , 1.2973790625068657 , 1.2970280068808187 , 1.2966751087734227 ,
                          1.2963203906293599 , 1.2959638750662528 , 1.2956055842619314 , 1.295245540390442 ,
                          1.2948837656258312 , 1.2945202821421453 , 1.294155112113431 , 1.2937882777137342 ,
                          1.2934198011171016 , 1.2930497044975797 , 1.292678010029215 , 1.2923047398860539 ,
                          1.2919299162421427 , 1.2915535612715279 , 1.291175697148256 , 1.2907963460463732 ,
                          1.2904155301399263 , 1.2900332716029614 , 1.289649592609525 , 1.2892645153336637 ,
                          1.2888780620575522 , 1.2884902551532955 , 1.288101116352612 , 1.2877106673345107 ,
                          1.2873189297780014 , 1.2869259253620935 , 1.2865316757657963 , 1.2861362026681191 ,
                          1.2857395277480712 , 1.2853416726846623 , 1.2849426591569015 , 1.2845425088437987 ,
                          1.2841412434243626 , 1.2837388845776032 , 1.2833354539825292 , 1.2829309733181509 ,
                          1.282525464263477 , 1.282118948497517 , 1.2817114476992806 , 1.281302983547777 ,
                          1.2808935777471486 , 1.280483252008391 , 1.2800720272188477 , 1.2796599240998547 ,
                          1.2792469633727486 , 1.2788331657588654 , 1.2784185519795417 , 1.2780031427561136 ,
                          1.2775869588099176 , 1.2771700208622898 , 1.2767523496345663 , 1.276333965848084 ,
                          1.2759148902241784 , 1.2754951434841864 , 1.2750747463494443 , 1.274653719541288 ,
                          1.2742320837810541 , 1.2738098597900789 , 1.2733870682896986 , 1.2729637300012493 ,
                          1.2725398656467188 , 1.2721154959206924 , 1.2716906405876798 , 1.2712653190372365 ,
                          1.2708395506589174 , 1.2704133548422778 , 1.269986750976873 , 1.269559758452258 ,
                          1.2691323966579882 , 1.2687046849836188 , 1.268276642818705 , 1.2678482895528016 ,
                          1.2674196445754646 , 1.2669907272762484 , 1.2665615570447089 , 1.2661321532704006 ,
                          1.2657025353428795 , 1.2652727226517 , 1.2648427345864177 , 1.264412590536588 ,
                          1.2639823098917657 , 1.2635519120078385 , 1.2631214153496457 , 1.2626908377015678 ,
                          1.2622601968372642 , 1.2618295105303947 , 1.261398796554619 , 1.2609680726835963 ,
                          1.2605373566909865 , 1.2601066663504492 , 1.259676019435644 , 1.2592454337202301 ,
                          1.258814926977868 , 1.2583845169822163 , 1.2579542215069353 , 1.2575240583256841 ,
                          1.2570940452121229 , 1.2566641999399109 , 1.2562345402827078 , 1.255805084014173 ,
                          1.2553758489079665 , 1.2549468528222276 , 1.2545181133309626 , 1.2540896464506535 ,
                          1.253661468027537 , 1.25323359390785 , 1.2528060399378298 , 1.2523788219637129 ,
                          1.2519519558317365 , 1.2515254573881371 , 1.2510993424791519 , 1.250673626951018 ,
                          1.2502483266499718 , 1.2498234574222506 , 1.2493990351140911 , 1.2489750755717306 ,
                          1.2485515946414056 , 1.2481286081693532 , 1.2477061320018101 , 1.2472841819850133 ,
                          1.2468627739652 , 1.2464419237886066 , 1.2460216473014707 , 1.2456019603500283 ,
                          1.2451828786518107 , 1.2447644166242198 , 1.2443465879456732 , 1.243929406286736 ,
                          1.2435128853179742 , 1.2430970387099536 , 1.242681880133239 , 1.2422674232583968 ,
                          1.241853681755992 , 1.2414406692965907 , 1.241028399550758 , 1.2406168861890599 ,
                          1.2402061428820619 , 1.2397961833003295 , 1.2393870211144284 , 1.238978669994924 ,
                          1.238571143612382 , 1.2381644556373683 , 1.237758619740448 , 1.237353649592187 ,
                          1.2369495588631507 , 1.236546361223905 , 1.2361440703449946 , 1.2357426994821594 ,
                          1.2353422604113105 , 1.234942764578498 , 1.2345442234297725 , 1.2341466484111845 ,
                          1.2337500509687835 , 1.2333544425486207 , 1.2329598345967459 , 1.2325662385592095 ,
                          1.232173665882062 , 1.2317821280113532 , 1.2313916363931339 , 1.2310022024734542 ,
                          1.2306138376983642 , 1.2302265535139147 , 1.2298403613661555 , 1.2294552727011374 ,
                          1.2290712989649102 , 1.2286884516035246 , 1.2283067420630305 , 1.2279261817894787 ,
                          1.227546782228919 , 1.227168554812027 , 1.2267915101269742 , 1.2264156575195166 ,
                          1.2260410062375473 , 1.225667565528958 , 1.2252953446416415 , 1.2249243528234894 ,
                          1.2245545993223947 , 1.2241860933862496 , 1.223818844262946 , 1.2234528612003768 ,
                          1.223088153446434 , 1.22272473024901 , 1.222362600855997 , 1.2220017745152874 ,
                          1.2216422604747736 , 1.2212840679823478 , 1.2209272062859025 , 1.2205716846333297 ,
                          1.220217512272522 , 1.2198646984513717 , 1.219513252417771 , 1.2191631834196122 ,
                          1.2188145006085862 , 1.2184672119278137 , 1.218121324523703 , 1.217776845531059 ,
                          1.2174337820846863 , 1.2170921413193903 , 1.2167519303699754 , 1.216413156371247 ,
                          1.2160758264580096 , 1.2157399477650686 , 1.2154055274272284 , 1.2150725725792944 ,
                          1.2147410903560711 , 1.214411087892364 , 1.2140825723229773 , 1.2137555507827165 ,
                          1.213430030406386 , 1.2131060183287914 , 1.212783521684737 , 1.212462547609028 ,
                          1.2121431032364693 , 1.211825195701866 , 1.2115088321400227 , 1.2111940187291936 ,
                          1.210880759986954 , 1.2105690613448823 , 1.2102589282578349 , 1.209950366180669 ,
                          1.2096433805682414 , 1.209337976875409 , 1.2090341605570287 , 1.2087319370679575 ,
                          1.2084313118630525 , 1.20813229039717 , 1.2078348781251678 , 1.207539080501902 ,
                          1.2072449029822299 , 1.2069523510210085 , 1.2066614300500595 , 1.206372144901078 ,
                          1.2060844997768392 , 1.2057984988516945 , 1.2055141462999952 , 1.2052314462960925 ,
                          1.204950403014338 , 1.2046710206290827 , 1.2043933033146779 , 1.204117255245475 ,
                          1.2038428805958254 , 1.20357018354008 , 1.2032991682525906 , 1.2030298389077083 ,
                          1.2027621996797841 , 1.20249625474317
                        };

//! Reference trace from a mathematica solution to a simple nose-hoover system
double q_reference[]  = { 1.6 , 1.6000093706007106 , 1.6000373852098038 ,
                          1.6000839718948376 , 1.600149029230002 , 1.6002324691819398 , 1.6003342047098827 ,
                          1.6004541487140338 , 1.6005922144718199 , 1.6007483170081624 , 1.6009223713878908 ,
                          1.6011142952517115 , 1.6013240069256256 , 1.6015514247222653 , 1.6017964681491301 ,
                          1.6020590593300759 , 1.6023391206770476 , 1.6026365746469493 , 1.6029513454947752 ,
                          1.6032833589004298 , 1.6036325405676766 , 1.6039988170796802 , 1.6043821172889294 ,
                          1.6047823704065514 , 1.605199505733803 , 1.6056334542899737 , 1.6060841486249413 ,
                          1.606551521341414 , 1.607035503509741 , 1.6075360296486498 , 1.6080530379677698 ,
                          1.6085864666788672 , 1.6091362539937077 , 1.6097023381240574 , 1.6102846572816827 ,
                          1.6108831496783491 , 1.6114977535348927 , 1.612128409168095 , 1.6127750614767342 ,
                          1.6134376559476538 , 1.614116138067697 , 1.6148104533237064 , 1.6155205472025258 ,
                          1.6162463651909984 , 1.6169878527759671 , 1.6177449555292622 , 1.6185176221031068 ,
                          1.6193058047491815 , 1.6201094558902562 , 1.620928527949101 , 1.6217629733484853 ,
                          1.6226127445111789 , 1.6234777938599516 , 1.6243580738175734 , 1.6252535370048777 ,
                          1.6261641394079467 , 1.6270898397538254 , 1.6280305968426654 , 1.628986369474619 ,
                          1.6299571164498379 , 1.630942796568474 , 1.6319433686306792 , 1.6329587914366055 ,
                          1.6339890241978607 , 1.6350340297634562 , 1.6360937728387306 , 1.637168218141071 ,
                          1.6382573303878643 , 1.6393610742964975 , 1.6404794145843575 , 1.6416123159688318 ,
                          1.6427597431673067 , 1.6439216616277776 , 1.6450980404606532 , 1.646288849893435 ,
                          1.647494060152938 , 1.648713641465978 , 1.64994756405937 , 1.651195798159929 ,
                          1.6524583139944704 , 1.653735081785897 , 1.6550260713193035 , 1.6563312567073312 ,
                          1.6576506142466816 , 1.6589841202340563 , 1.6603317509661564 , 1.6616934827396836 ,
                          1.6630692918513397 , 1.6644591545978256 , 1.6658630472758431 , 1.6672809461820939 ,
                          1.668712827613279 , 1.6701586678661005 , 1.6716184441103343 , 1.6730921378526298 ,
                          1.6745797318978732 , 1.676081209049553 , 1.6775965521111578 , 1.6791257438861764 ,
                          1.6806687671780973 , 1.682225604790409 , 1.6837962395266 , 1.6853806541901593 ,
                          1.686978831584575 , 1.688590754513336 , 1.6902164061520621 , 1.6918557732438637 ,
                          1.6935088444938702 , 1.6951756086259033 , 1.6968560543637847 , 1.6985501704313362 ,
                          1.7002579455523796 , 1.701979368450737 , 1.7037144278502299 , 1.70546311247468 ,
                          1.7072254110479095 , 1.7090013122937397 , 1.7107908050830412 , 1.712593881069459 ,
                          1.7144105343459268 , 1.7162407590837154 , 1.7180845494540955 , 1.7199418996283373 ,
                          1.7218128037777118 , 1.7236972560734893 , 1.7255952506869408 , 1.7275067817893366 ,
                          1.7294318435519473 , 1.7313704301460437 , 1.7333225357815911 , 1.7352881566447664 ,
                          1.737267291728804 , 1.7392599402332671 , 1.7412661013577204 , 1.7432857743017278 ,
                          1.7453189582648532 , 1.7473656524466605 , 1.7494258560467137 , 1.7514995682645769 ,
                          1.7535867882998142 , 1.7556875153519893 , 1.7578017486241109 , 1.7599294885708863 ,
                          1.7620707386210839 , 1.7642255026234297 , 1.766393784426649 , 1.7685755878794682 ,
                          1.7707709168306125 , 1.7729797751288077 , 1.7752021666227797 , 1.777438095161254 ,
                          1.7796875645929566 , 1.7819505787666132 , 1.7842271415309492 , 1.7865172574327517 ,
                          1.7888209339022894 , 1.7911381790941088 , 1.7934690011624403 , 1.7958134082615151 ,
                          1.798171408545564 , 1.8005430101688173 , 1.8029282212855062 , 1.8053270500498615 ,
                          1.8077395046161138 , 1.810165593138494 , 1.8126053237712327 , 1.8150587039201165 ,
                          1.8175257423692357 , 1.8200064519489558 , 1.8225008456221852 , 1.825008936351832 ,
                          1.8275307371008045 , 1.830066260832011 , 1.8326155205083603 , 1.83517852909276 ,
                          1.8377552995481188 , 1.840345844837345 , 1.842950177923347 , 1.8455683117690331 ,
                          1.8482002593373112 , 1.8508460335910903 , 1.8535056474932783 , 1.8561791140067836 ,
                          1.858866446131756 , 1.8615676589229777 , 1.864282770263722 , 1.8670117982117236 ,
                          1.8697547608247176 , 1.8725116761604388 , 1.8752825622766225 , 1.8780674372310033 ,
                          1.8808663190813164 , 1.8836792258852963 , 1.8865061757006787 , 1.889347186585198 ,
                          1.8922022765965896 , 1.895071463792588 , 1.8979547662309284 , 1.9008522019693457 ,
                          1.903763789065575 , 1.9066895455773512 , 1.9096294901075324 , 1.9125836442773931 ,
                          1.9155520306264586 , 1.9185346716873544 , 1.9215315899927055 , 1.9245428080751377 ,
                          1.9275683484672759 , 1.930608233701746 , 1.933662486311173 , 1.9367311288281828 ,
                          1.9398141837854 , 1.9429116737154506 , 1.9460236211509596 , 1.9491500486245528 ,
                          1.952290978668855 , 1.9554464338164923 , 1.9586164366000893 , 1.9618010095992364 ,
                          1.9650001771691437 , 1.968213965607837 , 1.971442401266162 , 1.9746855104949654 ,
                          1.9779433196450933 , 1.981215855067392 , 1.9845031431127074 , 1.9878052101318864 ,
                          1.9911220824757745 , 1.9944537864952183 , 1.997800348541064 , 2.001161794964158 ,
                          2.0045381521153463 , 2.007929446345475 , 2.0113357040053907 , 2.014756951445939 ,
                          2.018193215017967 , 2.0216445216115875 , 2.025110900329575 , 2.0285923806166286 ,
                          2.0320889919070138 , 2.035600763634994 , 2.039127725234835 , 2.0426699061408 ,
                          2.0462273357871537 , 2.049800043608161 , 2.053388059038086 , 2.056991411511193 ,
                          2.0606101304617472 , 2.064244245324012 , 2.0678937855322532 , 2.071558780520734 ,
                          2.0752392597237193 , 2.0789352525754734 , 2.082646788575449 , 2.0863738986796925 ,
                          2.0901166146521413 , 2.093874968153557 , 2.0976489908447005 , 2.101438714386333 ,
                          2.105244170439216 , 2.1090653906641097 , 2.1129024067217763 , 2.1167552502729765 ,
                          2.120623952978471 , 2.124508546499022 , 2.1284090624953897 , 2.1323255326283355 ,
                          2.136257988558621 , 2.140206461947007 , 2.1441709844542545 , 2.1481515877411246 ,
                          2.1521483040087035 , 2.1561611665542584 , 2.1601902082542863 , 2.1642354619719706 ,
                          2.1682969605704945 , 2.1723747369130426 , 2.176468823862798 , 2.1805792542829447 ,
                          2.1847060610366658 , 2.188849276987145 , 2.1930089349975663 , 2.1971850679311133 ,
                          2.20137770865097 , 2.205586890020319 , 2.209812644902345 , 2.2140550061602315 ,
                          2.2183140066571614 , 2.2225896793502815 , 2.226882058146171 , 2.2311911761718397 ,
                          2.2355170662577417 , 2.2398597612343285 , 2.244219293932052 , 2.2485956971813654 ,
                          2.2529890038127203 , 2.257399246656569 , 2.261826458543364 , 2.266270672303557 ,
                          2.270731920767601 , 2.2752102367659472 , 2.2797056531290494 , 2.284218202687358 ,
                          2.2887479182713273 , 2.2932948327114078 , 2.297858978838053 , 2.302440389991104 ,
                          2.307039098914977 , 2.311655136935026 , 2.31628853536271 , 2.320939325509486 ,
                          2.3256075386868127 , 2.330293206206148 , 2.334996359378949 , 2.339717029516675 ,
                          2.344455247930783 , 2.349211045932732 , 2.3539844548339786 , 2.3587755059459816 ,
                          2.363584230580199 , 2.3684106600480885 , 2.3732548256611086 , 2.3781167587307164 ,
                          2.3829964905489356 , 2.3878940513018305 , 2.392809469527359 , 2.397742773631331 ,
                          2.4026939920195574 , 2.4076631530978476 , 2.4126502852720115 , 2.41765541694786 ,
                          2.422678576531203 , 2.42771979242785 , 2.432779093043612 , 2.4378565067842985 ,
                          2.44295206205572 , 2.448065787263686 , 2.453197710814421 , 2.4583478611163296 ,
                          2.463516263803711 , 2.468702943218816 , 2.473907923703894 , 2.4791312296011956 ,
                          2.4843728852529723 , 2.4896329150014735 , 2.4949113431889502 , 2.500208194157653 ,
                          2.5055234922498317 , 2.5108572618077374 , 2.5162095271736202 , 2.5215803126897307 ,
                          2.5269696426983193 , 2.532377541608488 , 2.5378040314213712 , 2.543249130998177 ,
                          2.5487128590716646 , 2.554195234374591 , 2.559696275639714 , 2.565216001599791 ,
                          2.57075443098758 , 2.5763115825358383 , 2.5818874749773237 , 2.5874821270447943 ,
                          2.5930955574710075 , 2.598727784988721 , 2.604378828330692 , 2.6100487062716744 ,
                          2.6157374362391344 , 2.6214450305523918 , 2.627171500662011 , 2.6329168580185587 ,
                          2.6386811140725994 , 2.6444642802747 , 2.6502663680754255 , 2.6560873889253416 ,
                          2.661927354275014 , 2.6677862755750086 , 2.6736641642758907 , 2.6795610318282264 ,
                          2.685476889682581 , 2.6914117492895206 , 2.697365621691668 , 2.703338512210636 ,
                          2.7093304232328936 , 2.7153413571384477 , 2.721371316307306 , 2.727420303119475 ,
                          2.733488319954961 , 2.739575369193772 , 2.7456814532159144 , 2.7518065744013946 ,
                          2.757950735130221 , 2.764113937782399 , 2.770296184737936 , 2.776497478376839 ,
                          2.782717821060933 , 2.788957210697975 , 2.795215639112902 , 2.80149309777912 ,
                          2.807789578170036 , 2.814105071759055 , 2.820439570019585 , 2.826793064425032 ,
                          2.8331655464488015 , 2.839557007564301 , 2.8459674392449354 , 2.8523968329641134 ,
                          2.8588451801952397 , 2.8653124724117207 , 2.8717987011055994 , 2.878303855483077 ,
                          2.884827916068415 , 2.8913708616154303 , 2.8979326708779443 , 2.904513322609777 ,
                          2.9111127955647476 , 2.9177310684966757 , 2.924368120159381 , 2.9310239293066838 ,
                          2.937698474692404 , 2.9443917350703606 , 2.951103689194374 , 2.957834315818263 ,
                          2.9645835936958496 , 2.9713515008577644 , 2.978138006724619 , 2.9849430756901554 ,
                          2.9917666721039926 , 2.9986087603157485 , 3.005469304675042 , 3.012348269531492 ,
                          3.0192456192347183 , 3.0261613181343385 , 3.0330953305799717 , 3.0400476209212366 ,
                          3.047018153507753 , 3.0540068926891384 , 3.061013802815012 , 3.068038847520718 ,
                          3.0750819808267917 , 3.082143152351069 , 3.0892223120590168 , 3.096319409916103 ,
                          3.103434395887796 , 3.110567219939563 , 3.117717832036872 , 3.1248861821451905 ,
                          3.1320722202299867 , 3.139275896256729 , 3.146497160190883 , 3.153735961643498 ,
                          3.160992243248179 , 3.168265941339307 , 3.1755569920302857 , 3.1828653314345225 ,
                          3.190190895665423 , 3.1975336208363947 , 3.2048934430608416 , 3.212270298452171 ,
                          3.2196641231237892 , 3.2270748531891025 , 3.234502424761516 , 3.241946773190639 ,
                          3.2494078254796483 , 3.2568855035159414 , 3.2643797291143826 , 3.2718904240898357 ,
                          3.279417510257164 , 3.286960909431232 , 3.294520543426905 , 3.302096334059045 ,
                          3.3096882031425165 , 3.317296072492184 , 3.3249198639229114 , 3.332559497823413 ,
                          3.340214885303658 , 3.347885933762391 , 3.355572550588582 , 3.3632746431711964 ,
                          3.3709921188992036 , 3.378724885161571 , 3.3864728493472676 , 3.39423591884526 ,
                          3.402014001044517 , 3.4098070033340058 , 3.417614833102696 , 3.4254373954095456 ,
                          3.4332745858768967 , 3.4411262977462598 , 3.4489924242591887 , 3.456872858657238 ,
                          3.4647674941819604 , 3.4726762240749114 , 3.480598941577644 , 3.4885355399317137 ,
                          3.496485912378673 , 3.5044499521600763 , 3.5124275525106605 , 3.5204186032333524 ,
                          3.5284229854589286 , 3.5364405789963413 , 3.5444712636545423 , 3.552514919242484 ,
                          3.5605714255691168 , 3.5686406624433924 , 3.576722509674263 , 3.5848168470706816 ,
                          3.5929235544415974 , 3.6010425115959634 , 3.609173598283315 , 3.6173166897688382 ,
                          3.625471654190429 , 3.633638359098272 , 3.6418166720425504 , 3.6500064605734486 ,
                          3.65820759224115 , 3.6664199345958375 , 3.6746433551876954 , 3.682877721566907 ,
                          3.6911229012836566 , 3.699378761888127 , 3.707645170736996 , 3.7159219899984444 ,
                          3.72420907673996 , 3.7325062878791324 , 3.7408134803335544 , 3.7491305110208173 ,
                          3.757457236858513 , 3.7657935147642316 , 3.774139201655566 , 3.7824941544501063 ,
                          3.790858230065446 , 3.799231285419174 , 3.8076131770116195 , 3.81600375606504 ,
                          3.82440287086096 , 3.83281036973118 , 3.8412261010075035 , 3.849649913021731 ,
                          3.8580816541056655 , 3.866521172591109 , 3.874968316809862 , 3.8834229350937273 ,
                          3.891884875774507 , 3.900353987184003 , 3.9088301169540505 , 3.917313108133781 ,
                          3.9258028027840752 , 3.9342990430574165 , 3.9428016711062868 , 3.9513105290831696 ,
                          3.9598254591405473 , 3.9683463034309043 , 3.976872904106722 , 3.985405103320484 ,
                          3.993942743224673 , 4.002485665971773 , 4.011033712748871 , 4.019586721670661 ,
                          4.028144531334003 , 4.036706980391747 , 4.0452739074967425 , 4.053845151301841 ,
                          4.062420550459893 , 4.07099994362375 , 4.079583169446263 , 4.088170066580282 ,
                          4.096760473678656 , 4.105354229394238 , 4.113951171286478 , 4.122551136027224 ,
                          4.131153961603394 , 4.139759486014039 , 4.148367547258214 , 4.156977983334968 ,
                          4.1655906322433545 , 4.174205331982426 , 4.182821920551235 , 4.191440235948832 ,
                          4.200060116174271 , 4.208681399225248 , 4.217303922174441 , 4.2259275237459715 ,
                          4.234552044176978 , 4.243177323704597 , 4.251803202565963 , 4.260429520998214 ,
                          4.269056119238485 , 4.277682837523913 , 4.2863095160916345 , 4.294935995178786 ,
                          4.303562115022503 , 4.3121877158308175 , 4.320812637619705 , 4.329436724345739 ,
                          4.338059821298177 , 4.346681773766279 , 4.355302427039301 , 4.363921626406503 ,
                          4.372539217157143 , 4.381155044580478 , 4.389768953965766 , 4.398380790602268 ,
                          4.406990399779239 , 4.415597626721962 , 4.4242023178418615 , 4.43280432510813 ,
                          4.44140350148029 , 4.4499996999178695 , 4.458592773380391 , 4.467182574827383 ,
                          4.475768957218369 , 4.4843517735128735 , 4.492930876670422 , 4.501506119650542 ,
                          4.510077355412755 , 4.51864443687217 , 4.5272072200283136 , 4.53576556720844 ,
                          4.54431934136138 , 4.5528684054359605 , 4.561412622381013 , 4.5699518551453675 ,
                          4.578485966677853 , 4.5870148199273 , 4.595538277842537 , 4.604056203372394 ,
                          4.612568459465702 , 4.621074909179513 , 4.629575420819985 , 4.638069868947077 ,
                          4.64655812843764 , 4.6550400741685225 , 4.663515581016578 , 4.671984523858658 ,
                          4.680446777571613 , 4.688902217032293 , 4.69735071711755 , 4.705792152704236 ,
                          4.714226398669201 , 4.722653330362558 , 4.731072830482336 , 4.739484787219125 ,
                          4.747889088881633 , 4.75628562377858 , 4.764674280218675 , 4.7730549465106344 ,
                          4.781427510963172 , 4.789791861885003 , 4.798147887584838 , 4.806495476371393 ,
                          4.814834516553383 , 4.8231648975536885 , 4.831486517825585 , 4.8397992801287275 ,
                          4.848103087246526 , 4.856397841962387 , 4.864683447059722 , 4.872959805321937 ,
                          4.881226819532443 , 4.889484392474649 , 4.897732426931962 , 4.905970825687792 ,
                          4.914199491525548 , 4.922418329293905 , 4.930627253828378 , 4.938826182969695 ,
                          4.947015034559054 , 4.955193726437652 , 4.9633621764466875 , 4.971520302427359 ,
                          4.979668022220862 , 4.987805253668397 , 4.995931914611162 , 5.0040479228903525 ,
                          5.012153196349294 , 5.02024765614665 , 5.028331233449442 , 5.036403861305344 ,
                          5.044465472762035 , 5.052516000867188 , 5.060555378668483 , 5.068583539213595 ,
                          5.0766004155502 , 5.084605940725973 , 5.0926000477885935 , 5.100582669785736 ,
                          5.108553739802098 , 5.116513195622912 , 5.124460984207517 , 5.1323970535687815 ,
                          5.140321351719572 , 5.148233826672756 , 5.156134426441203 , 5.164023099037779 ,
                          5.171899792475352 , 5.179764454766788 , 5.1876170339249565 , 5.195457477962726 ,
                          5.2032857350408825 , 5.2111017592926805 , 5.2189055126155495 , 5.226696957414011 ,
                          5.234476056092581 , 5.24224277105578 , 5.249997064708127 , 5.25773889945414 ,
                          5.265468237698339 , 5.27318504184524 , 5.280889274299365 , 5.2885808974652315 ,
                          5.296259874108861 , 5.303926173919983 , 5.311579772662761 , 5.319220646292708 ,
                          5.326848770765335 , 5.334464122036157 , 5.342066676060685 , 5.349656408794433 ,
                          5.3572332961929146 , 5.36479731421164 , 5.3723484388061244 , 5.37988664593188 ,
                          5.387411912228544 , 5.394924221749987 , 5.402423562919911 , 5.409909924204228 ,
                          5.4173832940688476 , 5.424843660979682 , 5.43229101340264 , 5.439725339803635 ,
                          5.447146628648578 , 5.454554868403377 , 5.4619500475339455 , 5.469332154506193 ,
                          5.476701178888529 , 5.484057117629957 , 5.49139997053585 , 5.498729737406762 ,
                          5.506046418043245 , 5.5133500122458505 , 5.5206405198151325 , 5.527917940551645 ,
                          5.535182274255939 , 5.542433520728569 , 5.549671679770086 , 5.556896751181043 ,
                          5.564108736349216 , 5.5713076434948245 , 5.57849348250638 , 5.585666263267226 ,
                          5.592825995660704 , 5.599972689570157 , 5.607106354878925 , 5.614227001470352 ,
                          5.621334639227781 , 5.628429278034552 , 5.635510927774009 , 5.642579598329589 ,
                          5.649635301682319 , 5.6566780556538685 , 5.663707878933609 , 5.670724790210915 ,
                          5.677728808175161 , 5.68471995151572 , 5.691698238921965 , 5.698663689083271 ,
                          5.7056163206890105 , 5.712556152428559 , 5.719483202991289 , 5.7263974910816415 ,
                          5.73329903793295 , 5.740187869386089 , 5.747064011677021 , 5.753927491041703 ,
                          5.760778333716097 , 5.767616565936161 , 5.774442213937857 , 5.781255303957145 ,
                          5.788055862229983 , 5.794843914992333 , 5.801619488480154 , 5.808382608990215 ,
                          5.815133305591236 , 5.821871610718993 , 5.828597556946914 , 5.835311176848425 ,
                          5.842012502996954 , 5.8487015679659295 , 5.855378404328781 , 5.862043044658934 ,
                          5.868695521529818 , 5.8753358675148615 , 5.8819641151874915 , 5.8885802972570405 ,
                          5.895184449223943 , 5.901776608845541 , 5.908356813898179 , 5.914925102158201 ,
                          5.921481511401954 , 5.92802607940578 , 5.934558843946025 , 5.9410798427990335 ,
                          5.9475891137411505 , 5.954086694548723 , 5.96057262299809 , 5.967046936970481 ,
                          5.973509676509051 , 5.979960883684541 , 5.986400600645728 , 5.992828869541394 ,
                          5.999245732520316 , 6.005651231731276 , 6.012045409323052 , 6.018428307444424 ,
                          6.024799968244171 , 6.031160433871073 , 6.037509746473909 , 6.04384794820146 ,
                          6.050175081242035 , 6.056491189513285 , 6.062796318676689 , 6.0690905143708775 ,
                          6.075373822234487 , 6.0816462879061515 , 6.087907957024505 , 6.0941588752281834 ,
                          6.100399088155819 , 6.106628641446047 , 6.112847580737504 , 6.11905595166882 ,
                          6.125253799878632 , 6.131441171005573 , 6.137618111418948 , 6.14378466917696 ,
                          6.149940892372225 , 6.156086829096839 , 6.162222527442901 , 6.168348035502506 ,
                          6.174463401367751 , 6.180568673130735 , 6.1866638988835545 , 6.192749126718306 ,
                          6.198824404727088 , 6.204889781001999 , 6.210945303635133 , 6.2169910209541905 ,
                          6.223026982531221 , 6.22905323805655 , 6.235069837193272 , 6.241076829604482 ,
                          6.247074264953272 , 6.25306219290274 , 6.259040663115978 , 6.265009725256083 ,
                          6.270969428986146 , 6.276919823969264 , 6.282860959868531 , 6.288792886347041 ,
                          6.29471565311915 , 6.300629310621324 , 6.306533909421904 , 6.312429500010806 ,
                          6.318316132877942 , 6.324193858513229 , 6.33006272740658 , 6.335922790047911 ,
                          6.341774096927136 , 6.34761669853417 , 6.353450645358928 , 6.3592759878913245 ,
                          6.365092776621273 , 6.370901062040982 , 6.3767008949724415 , 6.382492326288951 ,
                          6.388275406690813 , 6.394050186878334 , 6.399816717551818 , 6.4055750494115715 ,
                          6.411325233157897 , 6.4170673194911 , 6.422801359111485 , 6.428527402719356 ,
                          6.4342455010150195 , 6.439955704698779 , 6.445658064470939 , 6.45135263113463 ,
                          6.45703945543837 , 6.462718587798929 , 6.468390078625437 , 6.474053978327021 ,
                          6.479710337312807 , 6.485359205991922 , 6.491000634773496 , 6.4966346740666525 ,
                          6.502261374280523 , 6.507880785824232 , 6.513492959106906 , 6.519097944537674 ,
                          6.524695792551382 , 6.530286553434842 , 6.53587027695467 , 6.541447012828603 ,
                          6.547016810774374 , 6.552579720509722 , 6.5581357917523775 , 6.563685074220079 ,
                          6.5692276176305615 , 6.5747634717015595 , 6.580292686150808 , 6.585815310696044 ,
                          6.591331395055001 , 6.59684098897857 , 6.602344142729373 , 6.607840905303128 ,
                          6.613331325202225 , 6.618815450929054 , 6.624293330986005 , 6.629765013875466 ,
                          6.6352305480998295 , 6.640689982161485 , 6.64614336456282 , 6.651590743806225 ,
                          6.657032168394092 , 6.6624676868288075 , 6.6678973476127625 , 6.673321199248348 ,
                          6.678739290237952 , 6.684151669083965 , 6.6895583842887785 , 6.694959484358462 ,
                          6.700355017599944 , 6.705745031064742 , 6.711129571479171 , 6.716508685569543 ,
                          6.721882420062171 , 6.727250821683368 , 6.7326139371594484 , 6.737971813216724 ,
                          6.743324496581507 , 6.748672033980111 , 6.75401447213885 , 6.759351857784037 ,
                          6.764684237641984 , 6.770011658439005 , 6.7753341669014135 , 6.780651809755521 ,
                          6.785964633727641 , 6.791272685544678 , 6.796576011654297 , 6.801874657117211 ,
                          6.807168666617173 , 6.812458084837936 , 6.817742956463252 , 6.823023326176875 ,
                          6.828299238662557 , 6.83357073860405 , 6.838837870685107 , 6.8441006795894825 ,
                          6.8493592100009275 , 6.8546135066031955 , 6.859863614080039 , 6.865109577115213 ,
                          6.870351440392467 , 6.875589248595554 , 6.880823046408229 , 6.886052878514243 ,
                          6.89127878930118 , 6.896500821692938 , 6.901719018177628 , 6.906933421243365 ,
                          6.912144073378259 , 6.917351017070422 , 6.922554294807968 , 6.927753949079008 ,
                          6.932950022371656 , 6.9381425571740225 , 6.943331595974221 , 6.948517181260363 ,
                          6.953699355520563 , 6.95887816124293 , 6.964053640915578 , 6.96922583702662 ,
                          6.974394792064168 , 6.979560548516334 , 6.984723148598106 , 6.989882633035616 ,
                          6.995039042052579 , 7.000192415872373 , 7.005342794718379 , 7.010490218813975 ,
                          7.0156347283825395 , 7.020776363647452 , 7.025915164832092 , 7.031051172159839 ,
                          7.036184425854071 , 7.041314966138167 , 7.046442833235505 , 7.051568067369466 ,
                          7.056690708763428 , 7.06181079764077 , 7.066928374224871 , 7.07204347873911 ,
                          7.0771561511759655 , 7.082266430060724 , 7.087374353344296 , 7.092479958976257 ,
                          7.09758328490618 , 7.10268436908364 , 7.107783249458212 , 7.112879963979469 ,
                          7.117974550596986 , 7.123067047260338 , 7.128157491919096 , 7.133245922522838 ,
                          7.138332377021137 , 7.1434168933635664 , 7.148499509499701 , 7.153580263379116 ,
                          7.158659192951384 , 7.163736336166081 , 7.168811730789619 , 7.17388541318167 ,
                          7.178957419051551 , 7.1840277841051465 , 7.189096544048335 , 7.194163734587 ,
                          7.1992293914270205 , 7.204293550274279 , 7.209356246834657 , 7.214417516814035 ,
                          7.219477395918295 , 7.224535919853317 , 7.229593124324983 , 7.234649045039174 ,
                          7.239703717701772 , 7.2447571780186575 , 7.249809461695713 , 7.254860604438817 ,
                          7.259910641816839 , 7.264959608083162 , 7.27000753676189 , 7.275054461369625 ,
                          7.280100415422968 , 7.285145432438519 , 7.290189545932879 , 7.295232789422647 ,
                          7.300275196424424 , 7.305316800454811 , 7.3103576350304085 , 7.315397733667817 ,
                          7.3204371298836355 , 7.3254758571944665 , 7.330513949116909 , 7.335551439167565 ,
                          7.340588360863033 , 7.345624747719915 , 7.350660633442099 , 7.355696051246884 ,
                          7.360731032427812 , 7.365765608148333 , 7.370799809571899 , 7.375833667861961 ,
                          7.380867214181968 , 7.38590047969537 , 7.390933495565618 , 7.395966292956164 ,
                          7.400998903030457 , 7.4060313569519485 , 7.411063685884088 , 7.416095920990327 ,
                          7.421128093434116 , 7.426160234378904 , 7.431192374988143 , 7.436224546425282 ,
                          7.441256779853774 , 7.446289106437067 , 7.451321557338614 , 7.456354163721863 ,
                          7.4613869566604745 , 7.466419965879732 , 7.47145322007856 , 7.476486747931336 ,
                          7.481520578112435 , 7.486554739296233 , 7.491589260157106 , 7.496624169369428 ,
                          7.501659495607579 , 7.50669526754593 , 7.511731513858859 , 7.516768263220741 ,
                          7.521805544305953 , 7.526843385788869 , 7.531881816343865 , 7.536920864645318 ,
                          7.541960559367603 , 7.547000929185097 , 7.552042002772172 , 7.557083808803208 ,
                          7.562126375952578 , 7.5671697328946586 , 7.572213908242504 , 7.577258929401917 ,
                          7.582304822686847 , 7.587351614373008 , 7.592399330736107 , 7.597447998051857 ,
                          7.602497642595967 , 7.607548290644149 , 7.6125999684721135 , 7.617652702355572 ,
                          7.622706518570233 , 7.627761443391809 , 7.63281750309601 , 7.637874723958547 ,
                          7.642933132255131 , 7.647992754261471 , 7.653053616253279 , 7.658115744506267 ,
                          7.663179165296143 , 7.668243904898621 , 7.673309989589408 , 7.678377445644218 ,
                          7.683446299300066 , 7.688516575729045 , 7.693588298937982 , 7.698661492875074 ,
                          7.703736181488515 , 7.708812388726502 , 7.71389013853723 , 7.718969454868895 ,
                          7.724050361669693 , 7.72913288288782 , 7.734217042471471 , 7.739302864368843 ,
                          7.74439037252813 , 7.749479590897529 , 7.754570543425237 , 7.759663254059447 ,
                          7.764757746748357 , 7.7698540454401614 , 7.774952174083057 , 7.7800521566252385 ,
                          7.785154017014904 , 7.790257779200246 , 7.795363467107092 , 7.8004711037303744 ,
                          7.805580710820965 , 7.810692310043156 , 7.815805923061242 , 7.820921571539515 ,
                          7.826039277142267 , 7.831159061533792 , 7.836280946378382 , 7.841404953340333 ,
                          7.846531104083933 , 7.851659420273478 , 7.85678992357326 , 7.861922635647572 ,
                          7.867057578160706 , 7.8721947727769574 , 7.877334241160617 , 7.882476004975979 ,
                          7.887620085887335 , 7.892766505558978 , 7.897915285655202 , 7.903066447840298 ,
                          7.9082200137673 , 7.913376004283872 , 7.918534438908833 , 7.923695337036762 ,
                          7.928858718062236 , 7.934024601379835 , 7.939193006384135 , 7.9443639524697165 ,
                          7.949537459031157 , 7.954713545463036 , 7.959892231159931 , 7.965073535516422 ,
                          7.970257477927086 , 7.9754440777865 , 7.980633354489246 , 7.985825327429899 ,
                          7.99102001600304 , 7.996217439603246 , 8.001417617625096 , 8.006620569463168 ,
                          8.011826314512042 , 8.017034872166295 , 8.022246261815953 , 8.027460502158752 ,
                          8.032677610475398 , 8.03789760387395 , 8.043120499462464 , 8.048346314349 ,
                          8.053575065641619 , 8.058806770448374 , 8.064041445877326 , 8.06927910903653 ,
                          8.074519777034048 , 8.079763466977935 , 8.08501019597625 , 8.090259981137054 ,
                          8.095512839568402 , 8.100768788378351 , 8.10602784467496 , 8.11129002556629 ,
                          8.116555348160396 , 8.121823829565338 , 8.127095486889171 , 8.132370337239957 ,
                          8.137648397724668 , 8.142929684858942 , 8.148214213652176 , 8.15350199888029 ,
                          8.158793055319212 , 8.164087397744856 , 8.169385040933149 , 8.174685999660007 ,
                          8.179990288701356 , 8.185297922833115 , 8.190608916831206 , 8.195923285471551 ,
                          8.20124104353007 , 8.206562205782683 , 8.211886787005316 , 8.217214801973887 ,
                          8.222546265464317 , 8.22788119225253 , 8.233219597114445 , 8.238561494825984 ,
                          8.243906900163068 , 8.249255827901619 , 8.254608292846878 , 8.259964311174707 ,
                          8.265323896438007 , 8.270687060628735 , 8.276053815738852 , 8.281424173760316 ,
                          8.286798146685085 , 8.292175746505118 , 8.297556985212376 , 8.302941874798815 ,
                          8.308330427256395 , 8.313722654577077 , 8.319118568752817 , 8.324518181775574 ,
                          8.329921505637309 , 8.335328552329978 , 8.34073933384554 , 8.346153862175957 ,
                          8.351572149313185 , 8.356994207249183 , 8.362420047975913 , 8.367849683485328 ,
                          8.373283125769392 , 8.378720386820063 , 8.384161478629297 , 8.389606413189055 ,
                          8.395055202491296 , 8.400507858527979 , 8.405964393291061 , 8.411424818774902 ,
                          8.41688914612669 , 8.42235738434428 , 8.427829542116058 , 8.433305628130409 ,
                          8.438785651075715 , 8.444269619640364 , 8.449757542512742 , 8.455249428381233 ,
                          8.460745285934221 , 8.466245123860094 , 8.471748950847235 , 8.477256775584031 ,
                          8.482768606758865 , 8.488284453060125 , 8.493804323176194 , 8.499328225795459 ,
                          8.504856169606303 , 8.510388163297113 , 8.515924215556273 , 8.52146433507217 ,
                          8.52700853053319 , 8.532556810627714 , 8.53810918404413 , 8.543665659470824 ,
                          8.54922624559618 , 8.554790951108584 , 8.56035978469642 , 8.565932755048076 ,
                          8.57150987084361 , 8.577091139546171 , 8.582676566547795 , 8.588266157054914 ,
                          8.593859916273965 , 8.599457849411381 , 8.605059961673597 , 8.610666258267045 ,
                          8.616276744398164 , 8.621891425273382 , 8.627510306099138 , 8.633133392081863 ,
                          8.638760688427995 , 8.644392200343965 , 8.65002793303621 , 8.65566789171116 ,
                          8.661312081575254 , 8.666960507834922 , 8.672613175696602 , 8.678270090366727 ,
                          8.68393125705173 , 8.689596680958047 , 8.695266367292112 , 8.700940321260356 ,
                          8.706618548069217 , 8.71230105292513 , 8.717987841034525 , 8.72367891760384 ,
                          8.729374287839507 , 8.735073956896802 , 8.740777928272257 , 8.746486203613145 ,
                          8.752198784476606 , 8.757915672419776 , 8.76363686899979 , 8.769362375773785 ,
                          8.775092194298898 , 8.780826326132269 , 8.786564772831031 , 8.792307535952322 ,
                          8.798054617053278 , 8.803806017691036 , 8.809561739422733 , 8.815321783805507 ,
                          8.821086152396493 , 8.826854846752829 , 8.832627868431652 , 8.838405218990097 ,
                          8.844186899985303 , 8.849972912974403 , 8.855763259514537 , 8.861557941162843 ,
                          8.867356959476453 , 8.873160316012509 , 8.878968012328144 , 8.884780049980497 ,
                          8.890596430526703 , 8.896417155523899 , 8.902242226373177 , 8.908071642372528 ,
                          8.91390540133804 , 8.919743501056214 , 8.92558593931355 , 8.931432713896548 ,
                          8.937283822591708 , 8.943139263185529 , 8.948999033464514 , 8.954863131215161 ,
                          8.960731554223972 , 8.96660430027745 , 8.972481367162088 , 8.978362752664392 ,
                          8.98424845457086 , 8.990138470667993 , 8.996032798742293 , 9.001931436580257 ,
                          9.007834381968387 , 9.013741632693183 , 9.019653186541147 , 9.025569041298777 ,
                          9.031489194752574 , 9.037413644689037 , 9.04334238889467 , 9.049275425155969 ,
                          9.055212751259438 , 9.061154364991575 , 9.06710026413888 , 9.073050446131239 ,
                          9.079004905966631 , 9.084963637628208 , 9.09092663509575 , 9.096893892349039 ,
                          9.102865403367854 , 9.108841162131975 , 9.114821162621187 , 9.120805398815266 ,
                          9.126793864693997 , 9.132786554237157 , 9.138783461424529 , 9.144784580235891 ,
                          9.150789904651027 , 9.156799428649718 , 9.162813146211741 , 9.16883105131688 ,
                          9.174853137944917 , 9.180879400075627 , 9.186909831688796 , 9.192944426764202 ,
                          9.198983179281628 , 9.205026083220853 , 9.211073132561658 , 9.217124321283825 ,
                          9.223179643367134 , 9.229239092791364 , 9.235302663536297 , 9.241370349581715 ,
                          9.2474421442251 , 9.253518038264493 , 9.259598021958595 , 9.265682085566347 ,
                          9.271770219346692 , 9.277862413558568 , 9.283958658460916 , 9.29005894431268 ,
                          9.296163261372797 , 9.302271599900212 , 9.308383950153864 , 9.314500302392693 ,
                          9.320620646875641 , 9.32674497386165 , 9.332873273609659 , 9.339005536378611 ,
                          9.345141752427445 , 9.351281912015104 , 9.357426005400526 , 9.363574022842656 ,
                          9.36972595460043 , 9.375881790932795 , 9.382041522098687 , 9.38820513835705 ,
                          9.394372629966822 , 9.400543987186948 , 9.406719200276365 , 9.412898259494016 ,
                          9.419081155095574 , 9.425267876207576 , 9.43145840977537 , 9.437652742575564 ,
                          9.443850861384764 , 9.450052752979582 , 9.456258404136626 , 9.462467801632505 ,
                          9.468680932243824 , 9.474897782747199 , 9.481118339919233 , 9.487342590536537 ,
                          9.493570521375718 , 9.499802119213387 , 9.506037370826151 , 9.51227626299062 ,
                          9.518518782483401 , 9.524764916081104 , 9.53101465056034 , 9.537267972697714 ,
                          9.543524869269836 , 9.549785327053314 , 9.55604933282476 , 9.562316873360778 ,
                          9.568587935437982 , 9.574862505832975 , 9.58114057132237 , 9.587422118682777 ,
                          9.5937071346908 , 9.599995606008811 , 9.606287515815387 , 9.612582847179116 ,
                          9.618881584178032 , 9.625183710890166 , 9.631489211393546 , 9.637798069766207 ,
                          9.644110270086177 , 9.650425796431486 , 9.656744632880168 , 9.663066763510253 ,
                          9.669392172399771 , 9.675720843626753 , 9.68205276126923 , 9.688387909405234 ,
                          9.694726272112796 , 9.701067833469944 , 9.707412577554711 , 9.71376048844513 ,
                          9.72011155021923 , 9.72646574695504 , 9.732823062724224 , 9.739183480819905 ,
                          9.745546983400569 , 9.751913552594285 , 9.758283170529127 , 9.764655819333173 ,
                          9.771031481134493 , 9.777410138061164 , 9.783791772241257 , 9.790176365802848 ,
                          9.796563900874014 , 9.802954359582824 , 9.809347724057353 , 9.815743976425678 ,
                          9.82214309881587 , 9.828545073356004 , 9.834949882174156 , 9.841357507398397 ,
                          9.847767931156804 , 9.85418113557745 , 9.860597102788407 , 9.86701581491629 ,
                          9.873437253446031 , 9.879861398882928 , 9.88628823176251 , 9.892717732620312 ,
                          9.899149881991862 , 9.905584660412698 , 9.912022048418347 , 9.918462026544344 ,
                          9.924904575326218 , 9.931349675299506 , 9.937797306999737 , 9.944247450962443 ,
                          9.95070008772316 , 9.957155197817416 , 9.963612761780746 , 9.97007276014868 ,
                          9.976535173456751 , 9.982999982240493 , 9.989467167035436 , 9.995936708377114 ,
                          10.002408586801057 , 10.00888278234767 , 10.015359274257683 , 10.021838041886294 ,
                          10.02831906458873 , 10.034802321720214 , 10.041287792635975 , 10.047775456691237 ,
                          10.054265293241228 , 10.060757281641173 , 10.067251401246303 , 10.073747631411837 ,
                          10.080245951493005 , 10.086746340845034 , 10.09324877882315 , 10.099753244782578 ,
                          10.106259718078546 , 10.112768178066279 , 10.119278604101007 , 10.12579097553795 ,
                          10.132305271732339 , 10.138821472039398 , 10.145339555448176 , 10.151859500355224 ,
                          10.158381285384472 , 10.164904889162496 , 10.171430290315872 , 10.17795746747117 ,
                          10.184486399254972 , 10.191017064293847 , 10.197549441214374 , 10.204083508643127 ,
                          10.210619245206681 , 10.21715662953161 , 10.223695640244493 , 10.230236255971901 ,
                          10.236778455340412 , 10.243322216976598 , 10.249867519507035 , 10.256414341558301 ,
                          10.262962661756967 , 10.269512458729611 , 10.27606371110281 , 10.282616397243952 ,
                          10.28917049515331 , 10.295725983192511 , 10.302282839732962 , 10.30884104314607 ,
                          10.315400571803247 , 10.3219614040759 , 10.32852351833544 , 10.33508689295327 ,
                          10.341651506300806 , 10.348217336749453 , 10.35478436267062 , 10.361352562435716 ,
                          10.367921914416149 , 10.37449239698333 , 10.381063988508666 , 10.387636667363568 ,
                          10.39421041191944 , 10.400785200547695 , 10.40736101161974 , 10.413937823506984 ,
                          10.420515614408098 , 10.427094362383707 , 10.43367404600565 , 10.440254643867556 ,
                          10.446836134563052 , 10.453418496685764 , 10.46000170882932 , 10.466585749587349 ,
                          10.473170597553477 , 10.479756231321332 , 10.486342629484543 , 10.492929770636735 ,
                          10.49951763337154 , 10.506106196282579 , 10.512695437963485 , 10.519285337007881 ,
                          10.525875872009399 , 10.532467021561667 , 10.539058764258305 , 10.54565107869295 ,
                          10.552243943459224 , 10.558837337043576 , 10.565431238010916 , 10.572025625600148 ,
                          10.578620479089055 , 10.58521577775543 , 10.591811500877059 , 10.598407627731726 ,
                          10.605004137597227 , 10.611601009751345 , 10.618198223471872 , 10.624795758036592 ,
                          10.631393592723299 , 10.637991706809776 , 10.644590079573812 , 10.651188690293198 ,
                          10.65778751824572 , 10.664386542709167 , 10.670985742961326 , 10.67758509827999 ,
                          10.684184587942939 , 10.690784191227971 , 10.697383887381745 , 10.703983656005459 ,
                          10.71058347741779 , 10.717183331983373 , 10.723783200066853 , 10.730383062032862 ,
                          10.736982898246046 , 10.743582689071038 , 10.750182414872482 , 10.756782056015016 ,
                          10.763381592863274 , 10.7699810057819 , 10.776580275135531 , 10.783179381288809 ,
                          10.789778304606369 , 10.796377025452852 , 10.802975524192899 , 10.809573781191142 ,
                          10.81617177681223 , 10.822769491420793 , 10.829366905367834 , 10.835963999216368 ,
                          10.842560754569314 , 10.849157153212886 , 10.855753176933309 , 10.862348807516799 ,
                          10.868944026749576 , 10.875538816417862 , 10.882133158307875 , 10.888727034205838 ,
                          10.895320425897967 , 10.901913315170487 , 10.908505683809613 , 10.91509751360157 ,
                          10.921688786332574 , 10.928279483788844 , 10.934869587756605 , 10.94145908002207 ,
                          10.948047942371467 , 10.954636156591011 , 10.961223704466812 , 10.967810567909442 ,
                          10.974396729990996 , 10.980982174215038 , 10.987566884085123 , 10.994150843104817 ,
                          11.000734034777679 , 11.00731644260727 , 11.013898050097149 , 11.02047884075088 ,
                          11.02705879807202 , 11.033637905564131 , 11.040216146730774 , 11.04679350507551 ,
                          11.053369964101902 , 11.059945507313504 , 11.066520118213885 , 11.0730937803066 ,
                          11.079666477095211 , 11.08623819208328 , 11.092808908774368 , 11.099378610733805 ,
                          11.10594728260824 , 11.112514909837628 , 11.119081477875858 , 11.125646972176828 ,
                          11.132211378194432 , 11.138774681382559 , 11.145336867195109 , 11.151897921085972 ,
                          11.158457828509045 , 11.16501657491822 , 11.17157414576739 , 11.17813052651045 ,
                          11.184685702601298 , 11.191239659493824 , 11.19779238264192 , 11.204343857499484 ,
                          11.210894069520407 , 11.217443004158586 , 11.223990646867911 , 11.2305369831192 ,
                          11.237081999239678 , 11.243625682720875 , 11.250168021133236 , 11.256709002047197 ,
                          11.263248613033202 , 11.269786841661688 , 11.276323675503102 , 11.282859102127878 ,
                          11.289393109106463 , 11.295925684009294 , 11.302456814406812 , 11.30898648786946 ,
                          11.315514691967675 , 11.3220414142719 , 11.328566642352577 , 11.335090363780148 ,
                          11.341612566125049 , 11.348133236957727 , 11.354652363848617 , 11.361169934369066 ,
                          11.367685936640921 , 11.37420036022162 , 11.380713194897446 , 11.38722443045467 ,
                          11.393734056679579 , 11.400242063358444 , 11.406748440277543 , 11.413253177223158 ,
                          11.419756263981565 , 11.426257690339042 , 11.432757446081865 , 11.439255520996316 ,
                          11.445751904868668 , 11.452246587485206 , 11.458739558632201 , 11.465230808095933 ,
                          11.471720325662684 , 11.478208101118724 , 11.484694124250339 , 11.491178384843801 ,
                          11.497660872960697 , 11.504141580142733 , 11.510620498426889 , 11.517097619850487 ,
                          11.523572936450853 , 11.530046440265314 , 11.536518123331197 , 11.542987977685828 ,
                          11.549455995366529 , 11.555922168410632 , 11.562386488855461 , 11.56884894873834 ,
                          11.575309540096601 , 11.581768254967562 , 11.588225085388554 , 11.594680023396903 ,
                          11.601133061029936 , 11.607584190324975 , 11.614033403319352 , 11.62048069205039 ,
                          11.626926048658229 , 11.633369466548215 , 11.639810939970458 , 11.646250463190047 ,
                          11.652688030472063 , 11.659123636081599 , 11.665557274283735 , 11.671988939343562 ,
                          11.678418625526161 , 11.68484632709662 , 11.691272038320026 , 11.697695753461462 ,
                          11.704117466786016 , 11.710537172558773 , 11.716954865044821 , 11.723370538509243 ,
                          11.729784187217126 , 11.736195805433557 , 11.74260538742362 , 11.749012927452402 ,
                          11.755418419806736 , 11.761821859669606 , 11.768223243385519 , 11.774622567372933 ,
                          11.781019828050301 , 11.787415021836072 , 11.793808145148708 , 11.800199194406655 ,
                          11.806588166028376 , 11.812975056432316 , 11.819359862036938 , 11.82574257926069 ,
                          11.832123204522027 , 11.838501734239404 , 11.844878164831275 , 11.851252492716094 ,
                          11.857624714312315 , 11.863994826038393 , 11.87036282431278 , 11.876728705553932 ,
                          11.88309246618067 , 11.889454103115462 , 11.895813614611637 , 11.902170999124175 ,
                          11.908526255108066 , 11.914879381018288 , 11.921230375309829 , 11.927579236437667 ,
                          11.933925962856792 , 11.940270553022183 , 11.946613005388823 , 11.9529533184117 ,
                          11.959291490545791 , 11.965627520246086 , 11.971961405967562 , 11.97829314616521 ,
                          11.984622739294007 , 11.990950183808941 , 11.997275478164994 , 12.003598620817147 ,
                          12.009919610220388 , 12.016238445043774 , 12.022555125213096 , 12.02886965107172 ,
                          12.03518202296095 , 12.04149224122208 , 12.04780030619641 , 12.05410621822524 ,
                          12.060409977649867 , 12.066711584811593 , 12.07301104005171 , 12.079308343711526 ,
                          12.085603496132332 , 12.091896497655432 , 12.098187348622123 , 12.1044760493737 ,
                          12.110762600251466 , 12.117047001596717 , 12.123329253750756 , 12.129609357054877 ,
                          12.13588731185038 , 12.142163118461243 , 12.148436777795613 , 12.154708291907516 ,
                          12.160977662946017 , 12.167244893060184 , 12.17350998439908 , 12.179772939111777 ,
                          12.186033759347332 , 12.192292447254818 , 12.198549004983297 , 12.204803434681835 ,
                          12.2110557384995 , 12.217305918585353 , 12.223553977088468 , 12.229799916157901 ,
                          12.236043737942728 , 12.242285444592005 , 12.248525038254805 , 12.254762521080192 ,
                          12.260997895217228 , 12.267231162814983 , 12.273462326022521 , 12.27969138698891 ,
                          12.285918348098077 , 12.292143212915345 , 12.298365985274568 , 12.304586669002994 ,
                          12.310805267927858 , 12.317021785876397 , 12.323236226675856 , 12.32944859415347 ,
                          12.335658892136482 , 12.341867124452127 , 12.34807329492765 , 12.354277407390285 ,
                          12.360479465667272 , 12.366679473585855 , 12.372877434973265 , 12.37907335365675 ,
                          12.385267233463544 , 12.391459078220889 , 12.397648891756022 , 12.403836677896185 ,
                          12.410022440468618 , 12.416206183300556 , 12.422387910219513 , 12.428567625520737 ,
                          12.434745334421795 , 12.440921042177113 , 12.447094754041125 , 12.453266475268256 ,
                          12.45943621111294 , 12.46560396682961 , 12.471769747672688 , 12.477933558896613 ,
                          12.484095405755808 , 12.490255293504708 , 12.496413227397738 , 12.502569212689336 ,
                          12.508723254633928 , 12.514875358485941 , 12.521025529499811 , 12.527173772929963 ,
                          12.533320094030833 , 12.539464498056844 , 12.545606990262433 , 12.551747575902025 ,
                          12.557886260230056 , 12.564023048524396 , 12.570157946726809 , 12.576290961291463 ,
                          12.582422098648115 , 12.58855136522654 , 12.594678767456502 , 12.60080431176777 ,
                          12.606928004590111 , 12.613049852353296 , 12.619169861487087 , 12.625288038421253 ,
                          12.631404389585567 , 12.63751892140979 , 12.643631640323692 , 12.64974255275704 ,
                          12.655851665139606 , 12.661958983901153 , 12.668064515471448 , 12.674168266280263 ,
                          12.680270242757361 , 12.686370451332513 , 12.692468898435484 , 12.698565590496045 ,
                          12.704660534034096 , 12.71075373626539 , 12.716845204569664 , 12.722934946306552 ,
                          12.729022968835702 , 12.735109279516747 , 12.741193885709336 , 12.747276794773102 ,
                          12.753358014067693 , 12.759437550952745 , 12.765515412787902 , 12.771591606932803 ,
                          12.777666140747089 , 12.783739021590401 , 12.78981025682238 , 12.795879853802667 ,
                          12.8019478198909 , 12.808014162446726 , 12.814078888829783 , 12.82014200639971 ,
                          12.826203522516149 , 12.832263444538741 , 12.838321779827128 , 12.844378536142166 ,
                          12.850433721882641 , 12.856487345020255 , 12.862539413516565 , 12.868589935333123 ,
                          12.874638918431483 , 12.880686370773203 , 12.886732300319833 , 12.892776715032936 ,
                          12.898819622874058 , 12.90486103180476 , 12.910900949786594 , 12.916939384781116 ,
                          12.922976344749882 , 12.929011837654445 , 12.935045871470372 , 12.941078454400445 ,
                          12.94710959478316 , 12.953139300951879 , 12.959167581239974 , 12.965194443980812 ,
                          12.971219897507758 , 12.977243950154184 , 12.983266610253455 , 12.989287886138944 ,
                          12.995307786144012 , 13.001326318602034 , 13.007343491846374 , 13.013359314210401 ,
                          13.019373794027484 , 13.025386939630987
                        };


//! Typedef'd NVEUpdator class factory
typedef boost::function<boost::shared_ptr<TwoStepNVT> (boost::shared_ptr<SystemDefinition> sysdef,
                                                 boost::shared_ptr<ParticleGroup> group,
                                                 boost::shared_ptr<ComputeThermo> thermo,
                                                 Scalar Q,
                                                 Scalar T)> twostepnvt_creator;

//! NVTUpdater creator
boost::shared_ptr<TwoStepNVT> base_class_nvt_creator(boost::shared_ptr<SystemDefinition> sysdef,
                                              boost::shared_ptr<ParticleGroup> group,
                                              boost::shared_ptr<ComputeThermo> thermo,
                                              Scalar Q,
                                              Scalar T)
    {
    boost::shared_ptr<VariantConst> T_variant(new VariantConst(T));
    return boost::shared_ptr<TwoStepNVT>(new TwoStepNVT(sysdef, group, thermo, Q, T_variant));
    }

#ifdef ENABLE_CUDA
//! NVTUpdaterGPU factory for the unit tests
boost::shared_ptr<TwoStepNVT> gpu_nvt_creator(boost::shared_ptr<SystemDefinition> sysdef,
                                       boost::shared_ptr<ParticleGroup> group,
                                       boost::shared_ptr<ComputeThermo> thermo,
                                       Scalar Q,
                                       Scalar T)
    {
    boost::shared_ptr<VariantConst> T_variant(new VariantConst(T));
    return boost::shared_ptr<TwoStepNVT>(new TwoStepNVTGPU(sysdef, group, thermo, Q, T_variant));
    }
#endif

//! Integrate 1 particle through time and compare to a mathematical solution
void nvt_updater_integrate_tests(twostepnvt_creator nvt_creator, boost::shared_ptr<ExecutionConfiguration> exec_conf)
    {
    // check that the nvt updater can actually integrate particle positions and velocities correctly
    // start with a 1 particle system to keep things simple: also put everything in a huge box so boundary conditions
    // don't come into play
    boost::shared_ptr<SystemDefinition> sysdef(new SystemDefinition(1, BoxDim(1000.0), 4, 0, 0, 0, 0, exec_conf));
    boost::shared_ptr<ParticleData> pdata = sysdef->getParticleData();
    boost::shared_ptr<ParticleSelector> selector_all(new ParticleSelectorTag(sysdef, 0, pdata->getN()-1));
    boost::shared_ptr<ParticleGroup> group_all(new ParticleGroup(sysdef, selector_all));


    // setup a simple initial state
    {
    ArrayHandle<Scalar4> h_pos(pdata->getPositions(), access_location::host, access_mode::readwrite);
    ArrayHandle<Scalar4> h_vel(pdata->getVelocities(), access_location::host, access_mode::readwrite);

    h_pos.data[0].x = 0.0;
    h_pos.data[0].y = 0.0;
    h_pos.data[0].z = Scalar(1.6);
    h_vel.data[0].x = 0.0;
    h_vel.data[0].y = 0.0;
    h_vel.data[0].z = 0.0;
    }

    Scalar deltaT = Scalar(0.005);
    Scalar Q = Scalar(2.0);
    Scalar T = Scalar(1.5/3.0);
    Scalar tau = sqrt(Q / (Scalar(3.0) * T));
    boost::shared_ptr<IntegratorTwoStep> nvt_up(new IntegratorTwoStep(sysdef, deltaT));
    boost::shared_ptr<ComputeThermo> thermo(new ComputeThermo(sysdef, group_all));
    boost::shared_ptr<TwoStepNVT> two_step_nvt = nvt_creator(sysdef, group_all, thermo, tau, T);
    two_step_nvt->setXi(Scalar(1.0));
    nvt_up->addIntegrationMethod(two_step_nvt);
    thermo->setNDOF(3);

    // see what happens with a constant force
    boost::shared_ptr<ConstForceCompute> fc1(new ConstForceCompute(sysdef, 0.0, 0.0, 0.75));
    nvt_up->addForceCompute(fc1);

    nvt_up->prepRun(0);

    // integrate through time and compare to the reference
    for (int i = 0; i < 2000; i++)
        {
        MY_BOOST_CHECK_CLOSE(pdata->getPosition(0).z, q_reference[i], tol);
        MY_BOOST_CHECK_CLOSE(pdata->getVelocity(0).z, p_reference[i], tol);

        nvt_up->update(i);
        }
    }

//! Compares the output from one NVEUpdater to another
void nvt_updater_compare_test(twostepnvt_creator nvt_creator1, twostepnvt_creator nvt_creator2, boost::shared_ptr<ExecutionConfiguration> exec_conf)
    {
    const unsigned int N = 1000;

    // create two identical random particle systems to simulate
    RandomInitializer rand_init(N, Scalar(0.2), Scalar(0.9), "A");
    boost::shared_ptr<SnapshotSystemData> snap;
    rand_init.setSeed(12345);
    snap = rand_init.getSnapshot();

    boost::shared_ptr<SystemDefinition> sysdef1(new SystemDefinition(snap, exec_conf));
    boost::shared_ptr<ParticleData> pdata1 = sysdef1->getParticleData();
    boost::shared_ptr<ParticleSelector> selector_all1(new ParticleSelectorTag(sysdef1, 0, pdata1->getN()-1));
    boost::shared_ptr<ParticleGroup> group_all1(new ParticleGroup(sysdef1, selector_all1));

    boost::shared_ptr<SystemDefinition> sysdef2(new SystemDefinition(snap, exec_conf));
    boost::shared_ptr<ParticleData> pdata2 = sysdef2->getParticleData();
    boost::shared_ptr<ParticleSelector> selector_all2(new ParticleSelectorTag(sysdef2, 0, pdata2->getN()-1));
    boost::shared_ptr<ParticleGroup> group_all2(new ParticleGroup(sysdef2, selector_all2));

    boost::shared_ptr<NeighborList> nlist1(new NeighborList(sysdef1, Scalar(3.0), Scalar(0.8)));
    boost::shared_ptr<NeighborList> nlist2(new NeighborList(sysdef2, Scalar(3.0), Scalar(0.8)));

    boost::shared_ptr<PotentialPairLJ> fc1(new PotentialPairLJ(sysdef1, nlist1));
    fc1->setRcut(0, 0, Scalar(3.0));
    boost::shared_ptr<PotentialPairLJ> fc2(new PotentialPairLJ(sysdef2, nlist2));
    fc2->setRcut(0, 0, Scalar(3.0));


    // setup some values for alpha and sigma
    Scalar epsilon = Scalar(1.0);
    Scalar sigma = Scalar(1.2);
    Scalar alpha = Scalar(0.45);
    Scalar lj1 = Scalar(4.0) * epsilon * pow(sigma,Scalar(12.0));
    Scalar lj2 = alpha * Scalar(4.0) * epsilon * pow(sigma,Scalar(6.0));

    // specify the force parameters
    fc1->setParams(0,0,make_scalar2(lj1,lj2));
    fc2->setParams(0,0,make_scalar2(lj1,lj2));

    boost::shared_ptr<IntegratorTwoStep> nvt1(new IntegratorTwoStep(sysdef1, Scalar(0.005)));
    boost::shared_ptr<ComputeThermo> thermo1(new ComputeThermo(sysdef1, group_all1));
    thermo1->setNDOF(3*N-3);
    boost::shared_ptr<TwoStepNVT> two_step_nvt1 = nvt_creator1(sysdef1, group_all1, thermo1, Scalar(0.5), Scalar(1.2));
    nvt1->addIntegrationMethod(two_step_nvt1);

    boost::shared_ptr<IntegratorTwoStep> nvt2(new IntegratorTwoStep(sysdef2, Scalar(0.005)));
    boost::shared_ptr<ComputeThermo> thermo2(new ComputeThermo(sysdef2, group_all2));
    thermo2->setNDOF(3*N-3);
    boost::shared_ptr<TwoStepNVT> two_step_nvt2 = nvt_creator2(sysdef2, group_all2, thermo2, Scalar(0.5), Scalar(1.2));
    nvt2->addIntegrationMethod(two_step_nvt2);

    nvt1->addForceCompute(fc1);
    nvt2->addForceCompute(fc2);

    nvt1->prepRun(0);
    nvt2->prepRun(0);

    // step for 3 time steps and verify that they are the same
    // we can't do much more because these things are chaotic and diverge quickly
    for (int i = 0; i < 5; i++)
        {
        {
        ArrayHandle<Scalar4> h_pos1(pdata1->getPositions(), access_location::host, access_mode::read);
        ArrayHandle<Scalar4> h_vel1(pdata1->getVelocities(), access_location::host, access_mode::read);
        ArrayHandle<Scalar3> h_accel1(pdata1->getAccelerations(), access_location::host, access_mode::read);
        ArrayHandle<Scalar4> h_pos2(pdata2->getPositions(), access_location::host, access_mode::read);
        ArrayHandle<Scalar4> h_vel2(pdata2->getVelocities(), access_location::host, access_mode::read);
        ArrayHandle<Scalar3> h_accel2(pdata2->getAccelerations(), access_location::host, access_mode::read);

        //cout << arrays1.x[100] << " " << arrays2.x[100] << endl;
        Scalar rough_tol = 2.0;

        // check position, velocity and acceleration
        for (unsigned int j = 0; j < N; j++)
            {
            MY_BOOST_CHECK_CLOSE(h_pos1.data[j].x, h_pos2.data[j].x, rough_tol);
            MY_BOOST_CHECK_CLOSE(h_pos1.data[j].y, h_pos2.data[j].y, rough_tol);
            MY_BOOST_CHECK_CLOSE(h_pos1.data[j].z, h_pos2.data[j].z, rough_tol);

            MY_BOOST_CHECK_CLOSE(h_vel1.data[j].x, h_vel2.data[j].x, rough_tol);
            MY_BOOST_CHECK_CLOSE(h_vel1.data[j].y, h_vel2.data[j].y, rough_tol);
            MY_BOOST_CHECK_CLOSE(h_vel1.data[j].z, h_vel2.data[j].z, rough_tol);

            MY_BOOST_CHECK_CLOSE(h_accel1.data[j].x, h_accel2.data[j].x, rough_tol);
            MY_BOOST_CHECK_CLOSE(h_accel1.data[j].y, h_accel2.data[j].y, rough_tol);
            MY_BOOST_CHECK_CLOSE(h_accel1.data[j].z, h_accel2.data[j].z, rough_tol);
            }

        }
        nvt1->update(i);
        nvt2->update(i);
        }
    }

//! Compares the output of NVTUpdater to a mathematica solution of a 1D problem
BOOST_AUTO_TEST_CASE( TwoStepNVT_mathematica_compare )
    {
    nvt_updater_integrate_tests(bind(base_class_nvt_creator, _1, _2, _3, _4, _5), boost::shared_ptr<ExecutionConfiguration>(new ExecutionConfiguration(ExecutionConfiguration::CPU)));
    }


#ifdef ENABLE_CUDA
//! Compares the output of NVTUpdaterGPU to a mathematica solution of a 1D problem
BOOST_AUTO_TEST_CASE( TwoStepNVTGPU_mathematica_compare )
    {
    nvt_updater_integrate_tests(bind(gpu_nvt_creator, _1, _2, _3, _4, _5), boost::shared_ptr<ExecutionConfiguration>(new ExecutionConfiguration(ExecutionConfiguration::GPU)));
    }

//! boost test case for comparing the GPU and CPU NVTUpdaters
BOOST_AUTO_TEST_CASE( TwoStepNVTGPU_comparison_tests)
    {
    twostepnvt_creator nvt_creator_gpu = bind(gpu_nvt_creator, _1, _2, _3, _4, _5);
    twostepnvt_creator nvt_creator = bind(base_class_nvt_creator, _1, _2, _3, _4, _5);
    nvt_updater_compare_test(nvt_creator, nvt_creator_gpu, boost::shared_ptr<ExecutionConfiguration>(new ExecutionConfiguration(ExecutionConfiguration::GPU)));
    }

#endif

#ifdef WIN32
#pragma warning( pop )
#endif
